from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Font import Font


class TrueTypeTable(Copyable):
    """
    TrueTypeTable - class to represent custom OpenType table
    """

    __slots__ = ["_parent", "_tag", "_value"]

    # Constructor

    def __init__(
        self,
        truetypetable_or_tag: TrueTypeTable | str | None = None,
        value: str | None = None,
        valuelen: int | None = None,
    ) -> None:
        """
        TrueTypeTable()              - generic constructor, creates an empty
                                       TrueTypeTable record
        TrueTypeTable(TrueTypeTable) - copy constructor
        TrueTypeTable(string tag)    - creates table, assigns 'tag' and empty value
        TrueTypeTable(string tag, string value)
                                     - creates TrueType table and assigns values to both
                                       attributes
        TrueTypeTable(string tag, string value, integer valuelen)
                                     - creates TrueType table and assigns values to both
                                       attributes.
        'value' may include zeroes

        Args:
            truetypetable_or_tag (TrueTypeTable | str | None, optional): _description_. Defaults to None.
            value (str | None, optional): _description_. Defaults to None.
            valuelen (int | None, optional): _description_. Defaults to None.
        """
        self._parent: Font | None = None
        self.tag = ""
        self.value = ""  # FIXME: str vs. bytes?
        if truetypetable_or_tag is not None:
            if isinstance(truetypetable_or_tag, TrueTypeTable):
                self._copy_constructor(truetypetable_or_tag)
            else:
                self.tag = truetypetable_or_tag
                if value is not None:
                    self.value = value
                    if valuelen is not None:
                        # FIXME: Just a hunch, take valuelen bytes from value
                        self.value = value[:valuelen]

    def __repr__(self) -> str:
        parent = self._parent or "orphan"
        return f"<TTTable: tag={self.tag}, {parent}>"

    # Attributes

    @property
    def parent(self) -> Font | None:
        return self._parent

    @property
    def tag(self) -> str:
        return self._tag

    @tag.setter
    def tag(self, value: str) -> None:
        self._tag = value

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value



from __future__ import annotations


def int32_to_unsigned(i: int) -> int:
    """
    Reinterpret a signed integer as unsigned
    """
    return int.from_bytes(i.to_bytes(4, signed=True), signed=False)


def uint32_to_signed(i: int) -> int:
    """
    Reinterpret an unsigned integer as signed
    """
    return int.from_bytes(i.to_bytes(4, signed=False), signed=True)



import unittest

from FL.helpers.nametables import StandardNametable

nt = StandardNametable()


class StandardNametableTest(unittest.TestCase):
    # Unicode to Name

    def test_uni_0(self):
        assert nt.get_name_for_unicode(0) == "NULL"

    def test_uni_65(self):
        assert nt.get_name_for_unicode(65) == "A"

    def test_uni_0x26b8(self):
        assert nt.get_name_for_unicode(0x26B8) == "uni26B8"

    def test_uni_0x1F60E(self):
        assert nt.get_name_for_unicode(0x1F60E) == "u1F60E"

    def test_uni_0x1F60EF(self):
        # longer
        assert nt.get_name_for_unicode(0x1F60EF) == "u1F60EF"

    # Name to Unicode

    def test_name_A(self):
        assert nt.get_unicodes_for_name("A") == [65]

    def test_name_union(self):
        assert nt.get_unicodes_for_name("union") == [0x222A]

    def test_name_dotnull(self):
        assert nt.get_unicodes_for_name(".null") == [0]

    def test_name_dotNUL(self):
        # not present
        assert nt.get_unicodes_for_name(".NUL") == [-1]

    def test_name_NUL(self):
        assert nt.get_unicodes_for_name("NUL") == [0]

    def test_name_NULL(self):
        assert nt.get_unicodes_for_name("NULL") == [0]

    def test_name_uni(self):
        assert nt.get_unicodes_for_name("uniFEFF") == [0xFEFF]

    def test_name_0x1F60E(self):
        assert nt.get_unicodes_for_name("u1F60E") == [0x1F60E]

    def test_name_0x1F60EF(self):
        assert nt.get_unicodes_for_name("u1F60EF") == [0x1F60EF]

    def test_name_0x1F60EJ(self):
        # nan
        assert nt.get_unicodes_for_name("u1F60EJ") == [-1]

    def test_name_carriagereturn(self):
        # multiple
        assert nt.get_unicodes_for_name("carriagereturn") == [13, 8629]



# Font formats

# FontLab VFB font
ftFONTLAB = 0

# PC Type 1 font (binary/PFB)
ftTYPE1 = 1

# PC MultipleMaster font (PFB)
ftTYPE1_MM = 20

# PC Type 1 font (ASCII/PFA)
ftTYPE1ASCII = 2

# PC MultipleMaster font (ASCII/PFA)
ftTYPE1ASCII_MM = 21

# PC TrueType/TT OpenType font (TTF)
ftTRUETYPE = 3

# PS OpenType (CFF-based) font (OTF)
ftOPENTYPE = 6

# Mac Type 1 font (generates suitcase and LWFN file, optionally AFM)
ftMACTYPE1 = 23

# Mac TrueType font (generates suitcase)
ftMACTRUETYPE = 30

# Mac TrueType font (generates suitcase with resources in data fork)
ftMACTRUETYPE_DFONT = 31



import unittest
from pathlib import Path

from FL.fake.Kerning import FakeKerning


class FakeKerningTests(unittest.TestCase):
    def test_instantiation_empty(self) -> None:
        k = FakeKerning()
        assert isinstance(k, FakeKerning)
        assert k._font is None
        assert k.kerning == {}
        assert k.flat_kerning == []
        assert k.classes == {"L": {}, "R": {}}
        assert k.classes_left == {}
        assert k.classes_right == {}

    def test_import_flc(self) -> None:
        k = FakeKerning()
        k.import_flc(Path(__file__).parent / "classes.flc")
        assert len(k.classes_left) == 2
        assert len(k.classes_right) == 1
        assert k.classes_left["A"].glyphs == [
            "Agrave",
            "Aacute",
            "Acircumflex",
            "Adieresis",
            "Aring",  # Aring has a trailing space in the FLC file!
        ]
        assert k.classes_left["Z"].glyphs == [
            "Zcaron",
            "Zacute",
            "Zdotaccent",
            "Zeta",
            "uni1E92",
            "uni1E94",
        ]
        assert k.classes_right["Z"].glyphs == [
            "Zcaron",
            "Zacute",
            "Zdotaccent",
            "Zeta",
            "uni1E92",
            "uni1E94",
        ]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from FL.fake.Kerning import FakeKerning


logger = logging.getLogger(__name__)


class KerningClass:
    """
    Object representing a kerning class.

    :param name: The class name.
    :type name:  str or None

    :param sides: The sides of the class. "", "L", "R", or "LR".
    :type sides:  str

    :param keyglyph: The name of the class key glyph.
    :type keyglyph:  str or None

    :param glyphs: The class member glyph names.
    :type glyphs:  list of str

    :param fromFLClass: A FontLab kerning class to import from.
    :type fromFLClass:  str or None

    :param parent: The :py:class:`fofa.kerning.KerningCollection`.
    :type parent:  str or None
    """

    def __init__(
        self,
        name: str | None = None,
        sides: str = "",
        keyglyph: str | None = None,
        glyphs: list[str] | None = None,
        fromFLClass: str | None = None,
        parent: FakeKerning | None = None,
    ) -> None:
        self.name = name
        self.sides = sides  # "", "L", "R", "LR"
        self._keyglyph = keyglyph
        self.glyphs = glyphs or []
        self._parent = parent
        if self.keyglyph in self.glyphs:
            self.glyphs.remove(self.keyglyph)
        if fromFLClass:
            self.importFromFontLabClass(fromFLClass)

    def __len__(self) -> int:
        length = len(self.glyphs)
        if self.keyglyph:
            length += 1
        return length

    def __repr__(self) -> str:
        result = f"Kerning class {self.name} ({self.sides})\n"
        result += "  Key: %s\n" % self.keyglyph
        result += "  Glyphs: %s\n\n" % self.glyphs
        return result

    @classmethod
    def fromFontLabClass(cls, flClass: str) -> KerningClass:
        kc = cls()
        kc.importFromFontLabClass(flClass)
        return kc

    @property
    def keyglyph(self) -> str | None:
        """
        Return or set the keyglyph to a glyph name. Setting the keyglyph
        removes that glyph name from the members list.
        """
        return self._keyglyph

    @keyglyph.setter
    def keyglyph(self, value: str | None) -> None:
        if value is None:
            # Key glyph can not be none
            raise ValueError

        if self._keyglyph:
            self.glyphs.append(self._keyglyph)
        self._keyglyph = value
        if self._keyglyph in self.glyphs:
            self.glyphs.remove(self._keyglyph)

    def importFromMMClass(self, mmClass: str) -> None:
        """
        Import the class from a MetricsMachine class.

        :param mmClass: The MM class as string.
        :type mmClass:  str or None
        """
        # Check lfFontTools.kerningClass if you need it
        raise NotImplementedError

    def importFromFontLabClass(self, flClass: str) -> None:
        """
        Builds a new kerning class from an fl.font.classes element string, e.g.
        "_LAT_a_LEFT: a' adieresis".

        Args:
            flClass (str): The FontLab class string

        Raises:
            ValueError: If there is a duplicate key glyph.
        """
        self.name = None
        self.glyphs = []
        self._keyglyph = None

        parts = flClass.split()
        self.name = parts[0][:-1]
        for p in parts[1:]:
            if len(p) > 0:  # ignore multiple spaces in FL class definitions
                if p[-1] == "'":
                    if self._keyglyph is None:
                        self.keyglyph = p[:-1]
                    else:
                        logger.warning(f"Duplicate key glyph in class {self.name}")
                        raise ValueError
                else:
                    self.glyphs.append(p)
        side = self.name.rsplit("_", 1)[-1]
        if side in ("LEFT", "1st", "1"):
            self.sides = "L"
        elif side in ("RIGHT", "2nd", "2"):
            self.sides = "R"
        else:
            self.sides = "LR"
        if self._keyglyph is None:
            # Fallback to first member glyph
            if self.glyphs:
                self.keyglyph = self.glyphs[0]
            else:
                logger.warning("Empty class: %s" % self.name)
                logger.warning("             %s" % flClass)

    def getFontLabExternalClassCode(self) -> str:
        """
        External class code is the format written by FontLab when saving classes from
        the Classes panel.
        """
        if self.sides == "L":
            sideName = "_LEFT"
        elif self.sides == "R":
            sideName = "_RIGHT"
        else:
            sideName = ""
            self.sides = "LR"

        result = f"%%CLASS _{self.name}{sideName}\n"
        result += f"%%GLYPHS {self.keyglyph}' {' '.join(self.glyphs)}\n"
        result += "%%%%KERNING %s 0\n%%%%END\n" % (self.sides)

        return result



from __future__ import annotations

from FL.helpers.tth import TT_COMMANDS


class TTHCommand:
    """
    TTHCommand - class to represent visual TrueType instruction
    """

    __slots__ = ["_code", "_params"]

    # Constructor

    def __init__(self, code: int | None = None, *args: int) -> None:
        """
        Creates command, assigns command code and one or more parameters

        Args:
            code (int): The command code.
            *args (int): Zero or more parameters.

        See FL.helpers.tth.TT_COMMANDS for valid command codes and their matching
        parameters.
        """
        if code is None or len(args) >= 5:
            # More than 4 params will cause a RuntimeError, even though it should be
            # at > 5 according to the message
            raise RuntimeError(
                "Incorrect # of args to: "
                "TTHCommand(Int Code[,<up to 5 Int parameters>])"
            )

        if 1 <= code <= 23:
            self._code = code
            self._params = list(args)
            # Invalid codes will make FontLab 5 enter an infinite loop
            # Here, we just throw an exception
            if code not in TT_COMMANDS:
                raise ValueError(
                    "Codes 9, 10, 11, 12, 15, 16, 17, 18, 19 "
                    "will send FontLab into an infinite loop."
                )
        else:
            raise RuntimeError(
                "Incorrect instruction code. "
                "Value must be within the range from 1 to 23"
            )
        if len(args) == 0:
            # If there are too few args, the params are initialized with garbage ...
            self._params = [1632632832]

    def __repr__(self) -> str:
        cmd = TT_COMMANDS.get(self.code, {}).get("name", "NOTHING").upper()
        return f"<TTHCommand: {cmd}{tuple(self.params)}>"

    # Attributes

    @property
    def code(self) -> int:
        return self._code

    @code.setter
    def code(self, value: int) -> None:
        raise RuntimeError("class TTHCommand has no attribute code")

    @property
    def params(self) -> list[int]:
        return self._params

    @params.setter
    def params(self, value: list[int]) -> None:
        raise RuntimeError(
            'Attempt to write read only attribute "params" of class TTHCommand'
        )

    # Operations: none

    # Methods: none



from __future__ import annotations

from typing import Any, Iterable


def adjust_list(seq: list[Any], new_length: int, value: Any = None) -> None:
    """
    Adjust the length of a list in place.

    Args:
        seq (list[Any]): The list to be adjusted
        new_length (int): The desired length
        value (Any): The default value for added list elements
    """
    old_length = len(seq)
    diff = new_length - old_length
    if diff == 0:
        return

    if diff > 0:
        # Add elements
        if value is None and seq:
            value = seq[0]
        seq.extend([value] * diff)
    else:
        # Take away elements
        for _ in range(abs(diff)):
            seq.pop()


class FLList(list[Any]):
    def __init__(self, iterable: Iterable[Any] | None = None) -> None:
        if iterable is None:
            iterable = []
        super().__init__(iterable)

    # FontLab-specific

    def clean(self) -> None:
        super().clear()



import unittest
from pathlib import Path

from FL import Options


def get_reg_path(filename):
    return Path(__file__).parent.parent / "data" / filename


class OptionsTests(unittest.TestCase):
    def test_parse_minimal_ok(self):
        o = Options()

        # Set some values different from the defaults
        o.CacheTTPath = r"C:\Programme\cachett.exe"
        o.ExpandKernCodepage = "Foo"
        o.ExpandKernCount = 1000

        o.fake_load_regfile(get_reg_path("minimal.reg"))

        # Values should have been overwritten
        assert o.CacheTTPath == ""
        assert o.ExpandKernCodepage == "MS Windows 1252 Western (ANSI)"
        assert o.ExpandKernCount == 10919

    def test_parse_full_ok(self):
        o = Options()

        # Set some values different from the defaults
        o.CacheTTPath = r"C:\Programme\cachett.exe"
        o.ExpandKernCodepage = "Foo"
        o.ExpandKernCount = 1000

        o.fake_load_regfile(get_reg_path("Untitled.reg"))

        # Values should have been overwritten
        assert o.CacheTTPath == ""
        assert o.ExpandKernCodepage == "MS Windows 1252 Western (ANSI)"
        assert o.ExpandKernCount == 10919



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from vfbLib.parsers.text import OpenTypeStringParser

from FL.fake.Base import Copyable
from FL.fake.Kerning import FakeKerning
from FL.fake.mixins import GuideMixin, GuidePropertiesMixin
from FL.objects.Feature import Feature

if TYPE_CHECKING:
    from FL.objects.Uni import Uni


class FakeFont(Copyable, GuideMixin, GuidePropertiesMixin):
    __slots__ = [
        "_fake_binaries",
        "_fake_kerning",
        "_file_name",
        "_selection",
        "fake_sparse_json",
        "fake_vfb_object",
    ]

    def __init__(self) -> None:
        # Additions for FakeLab

        self._fake_binaries: dict[str, str] = {}
        self._fake_kerning = FakeKerning(self)
        self.fake_sparse_json = True
        self.fake_deselect_all()

    # Additional properties for FakeLab

    @property
    def fake_kerning(self) -> FakeKerning:
        """
        Returns the `FL.fake.FakeKerning` object, which can be used to manipulate the
        font's kerning data.
        """
        return self._fake_kerning

    # Additions for FakeLab

    def fake_binary_get(self, fontType: int) -> bytes:
        binary_path = self._fake_binaries[str(fontType)]
        with open(binary_path, "rb") as f:
            binary = f.read()
        return binary

    def fake_binary_from_path(self, fontType: int, file_path: str) -> None:
        """
        Assign a binary file from a path. This will be used to fake the
        FakeLab.GenerateFont() method.
        """
        # Convert key to str because JSON needs it
        self._fake_binaries[str(fontType)] = file_path

    def fake_update(self) -> None:
        """
        Is called from FontLab.UpdateFont()
        """
        for index, glyph in enumerate(self.glyphs):
            glyph.fake_update(self, index)

    def fake_deselect_all(self) -> None:
        """
        Deselect all glyphs. Is called from FontLab.Unselect().
        """
        self._selection: set[int] = set()

    def fake_select(self, gid: str | Uni | int, value: bool | None = None) -> None:
        """
        Change selection status for glyph_index.
        >>> f = Font()
        >>> f.fake_select(1, False)
        >>> print(f._selection)
        set()
        >>> f.fake_select(1, True)
        >>> print(f._selection)
        {1}
        >>> f.fake_select(3, True)
        >>> print(f._selection)
        {1, 3}
        >>> f.fake_select(2, False)
        >>> print(f._selection)
        {1, 3}
        >>> f.fake_select(1, False)
        >>> print(f._selection)
        {3}
        """
        if isinstance(gid, int):
            glyph_index = gid
        elif isinstance(gid, Uni) or isinstance(gid, str):
            glyph_index = self.FindGlyph(gid)
        if glyph_index > -1:
            if value:
                self._selection |= {glyph_index}
            else:
                self._selection -= {glyph_index}

    def fake_set_class_flags(self, flags: list[str]) -> None:
        """
        Set the kerning class flags from a list of str ("L", "R", "LR", ...)
        """
        # FIXME: In the vfb, the flags are stored as tuples of two values. The second
        # value is 0, it's not clear what it represents. Maybe the width flag for
        # metrics classes?
        for i, f in enumerate(flags):
            self.SetClassFlags(i, "L" in f, "R" in f)

    def _set_file_name(self, filename: str | Path | None) -> None:
        """
        Make sure the file name (actually, the path) is stored as Path
        """
        if filename is None:
            self._file_name = None
            return

        self._file_name = Path(filename) if not isinstance(filename, Path) else filename

    def fake_deserialize_axis(self, data: str) -> None:
        self.axis.append((data, data[:2], data))

    def fake_serialize_axis(self) -> list[str]:
        names = []
        for axis in self.axis:
            long, _, _ = axis
            names.append(long)
        return names

    def fake_deserialize_features(self, features: list[str]) -> None:
        self._features.clean()
        features_dict = OpenTypeStringParser.build_fea_dict(features)
        prefix = features_dict.get("prefix")
        if prefix:
            self.ot_classes = "\n".join(prefix)
        for feature_dict in features_dict.get("features", []):
            feature = Feature(feature_dict["tag"], "\n".join(feature_dict["code"]))
            self.features.append(feature)

    def fake_serialize_features(self) -> list[str]:
        fea = []
        if self.ot_classes:
            fea.extend(self.ot_classes.splitlines())
            # FIXME: Do we need empty lines as separator?
            fea.append("\n")
        for feature in self.features:
            fea.extend(feature.value.splitlines())
            fea.append("")
        return fea



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from FL import ftFONTLAB, ftOPENTYPE, ftTRUETYPE
from FL.objects.Font import Font
from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects.Canvas import Canvas
    from FL.objects.Glyph import Glyph
    from FL.objects.Rect import Rect
    from FL.objects.Uni import Uni


class FakeLab:
    """
    The FontLab program interface

    The main class. It is used via the pre-instantiated object `fl`.

    There is no explicit constructor for this class, use pre-initialized object 'fl'

    Always use "from FL import all" at the beginning of your macro.
    """

    __slots__ = [
        "_font",
        "_ifont",
        "_fonts",
        "ifontslist",
        "_glyph",
        "iglyph",
        "_tobject",
        "_iobject",
        "delta",
        "scale",
        "_tablet_active",
        "_tablet_pressure",
        "preview",
        "_layer",
    ]

    # Constructor

    def __init__(self) -> None:
        self._font: Font | None = None

        # (integer)  - index of currently active font
        self._ifont = -1

        self._fonts: list[Font] = []

        # (integer) - index of currently selected font in the fonts list panel
        self.ifontslist = 0

        self._glyph: Glyph | None = None

        # (integer)  - index of currently active glyph
        self.iglyph = -1

        # (integer)  - read-only - type of the currently selected object in the
        # Glyph Window
        self._tobject: int = -1

        # (integer)  - read-only - index of currently selected object in the
        # Glyph Window
        self._iobject = None

        # (<a href="Point.xml.html">Point</a>)  - delta value of current
        # coordinate translation in the active Glyph Window
        self.delta: Point | None = None

        # (<a href="Point.xml.html">Point</a>(float))      - scale value of
        # current coordinate translation in the active Glyph Window
        self.scale: Point | None = None

        # (boolean)   - True if tablet is present and active
        self._tablet_active = False

        # (integer) - current tablet's pen pressure
        self._tablet_pressure = 0

        # (string)          - contents of the preview panel <font color="red">
        # (not reported by docstring)</font>
        self.preview = ""

        self._layer: int | None = None

    # Operations

    def __len__(self) -> int:
        """
        returns number of opened fonts
        """
        return len(self._fonts)

    def __getitem__(self, index: int) -> Font:
        """
        returns Font by index
        """
        return self._fonts[index]

    # Attributes

    @property
    def font(self) -> Font | None:
        """
        Return the currently active font or None.
        """
        return self._font

    @property
    def ifont(self) -> int:
        """
        Return the index of the currently active font, or -1 if there is no font
        """
        return self._ifont

    @ifont.setter
    def ifont(self, value: int) -> None:
        self._ifont = value
        if self._ifont == -1:
            self._font = None
        else:
            self._font = self._fonts[self._ifont]

    @property
    def glyph(self) -> Glyph | None:
        """
        Return the currently active glyph in Font, Glyph or Metrics windows.
        """
        return self._glyph

    @property
    def tobject(self) -> int:
        """
        Return the type of the currently selected object in the Glyph Window as int.
        """
        # TODO: Actual return values

        # If there is no glyph window, raise:
        raise SystemError

    @property
    def iobject(self) -> int:
        """
        Return the index of currently selected object in the Glyph Window.
        """
        # TODO: Actual return values

        # If there is no glyph window, raise:
        raise SystemError

    @property
    def mainwindow(self) -> int:
        """
        Return a reference to FontLab's main window
        """
        return 0

    @property
    def path(self) -> str:
        """
        full path to directory where running application is located
        """
        return "/Library/Application Support/FontLab/Studio 5"

    @property
    def filename(self) -> str:
        """
        application filename
        """
        return "FontLab Studio 5 5730.app"

    @property
    def version(self) -> str:
        """
        application version
        """
        return "5.1.6/Mac(Build 7030)"

    @property
    def productnumber(self) -> int:
        """
        product number
        """
        return 0

    @property
    def serialnumber(self) -> str:
        """
        serial number as appears in the About window
        """
        return ""

    @property
    def username(self) -> str:
        """
        user name as appears in the About window
        """
        return "FontLab User"

    @property
    def count(self) -> int:
        """
        number of opened fonts (fast operation)
        """
        return self.__len__()

    @property
    def count_selected(self) -> int:
        """
        number of the selected glyphs in the Font Window (fast operation)
        """
        # FIXME
        return 0

    @property
    def window(self) -> int:
        """
        reference to the currently active Glyph, Font or Metrics window
        """
        # FIXME
        return 0

    # Properties not reported in the docs

    @property
    def layer(self) -> int | None:
        """
        Return the index of the currently active layer (master) in Glyph or Metrics
        windows.
        """
        return self._layer

    # Methods

    def Close(self, fontindex: int | None = None) -> None:
        """
        () | (fontindex)      - closes the current or 'fontindex' font
        """
        if fontindex is None:
            del self._fonts[self._ifont]
        else:
            del self._fonts[fontindex]
        if self.count == 0:
            self.ifont = -1
        else:
            self.ifont = self.count - 1

    def Open(self, filename: str, addtolist: bool = True) -> None:
        """
        Open the font from file using current opening options. If `addtolist` is True,
        the font is added to FontLab's font list. The font is shown in a window.

        Args:
            filename (str): _description_
            addtolist (bool, optional): _description_. Defaults to True.

        `addtolist` seems to be ignored; the font window is always opened.
        """
        font = Font()
        font.Open(filename)
        self.Add(font)

    def Save(
        self, filename_or_fontindex: str | int, filename: str | None = None
    ) -> None:
        """
        (string filename) | (int fontindex, string filename)

        Save the current or selected font using standard FontLab's Save
        routine.
        """
        if isinstance(filename_or_fontindex, int):
            # Save the font fontindex
            fontindex = filename_or_fontindex
            if 0 <= fontindex < self.count:
                if filename is None:
                    raise RuntimeError
                self._fonts[fontindex].Save(filename)
        else:
            # Save the current font
            if self.font is not None:
                filename = filename_or_fontindex
                self.font.Save(filename)

    def GenerateFont(self, *args: Any) -> int:
        """
        Generate a Font.

        Args:
            fontindex_or_fonttype (int): The index of the font to be generated, or the
                font type, see below.
            fonttype (int): The font type, see below.
            filename (str): The path for the font to be generated.

        Returns:
            int: 0 on success, -1 if the font was not generated.

        Available font types:

        - ftFONTLAB            - FontLab VFB font
        - ftTYPE1              - PC Type 1 font (binary/PFB)
        - ftTYPE1_MM           - PC MultipleMaster font (PFB)
        - ftTYPE1ASCII         - PC Type 1 font (ASCII/PFA)
        - ftTYPE1ASCII_MM      - PC MultipleMaster font (ASCII/PFA)
        - ftTRUETYPE           - PC TrueType/TT OpenType font (TTF)
        - ftOPENTYPE           - PS OpenType (CFF-based) font (OTF)
        - ftMACTYPE1           - Mac Type 1 font (generates suitcase and LWFN file,
                                 optionally AFM)
        - ftMACTRUETYPE        - Mac TrueType font (generates suitcase)
        - ftMACTRUETYPE_DFONT  - Mac TrueType font (generates suitcase with resources in
                                 data fork)
        """
        try:
            from ufo2ft import compileOTF, compileTTF
            from vfbLib.ufo.builder import VfbToUfoBuilder

            can_generate = True
        except ImportError:
            can_generate = False

        if len(args) == 3:
            fontindex, fonttype, filename = args
        elif len(args) == 2:
            fonttype, filename = args
            fontindex = self.ifont
        else:
            raise RuntimeError(
                "Incorrect # of args to:\nFontLab.GenerateFont([Int fontindex,] Int "
                "format, String filename)"
            )

        # An invalid font index was requested, or there is no open font
        if fontindex == -1:
            raise RuntimeError

        font = self._fonts[fontindex]

        if can_generate:
            # At the moment, we jump through some hoops. The font must have been opened
            # from a VFB file, and we need to build an UFO in memory to generate.
            builder = VfbToUfoBuilder(
                font.fake_vfb_object, add_kerning_groups=True, move_groups=True
            )
            # TODO: What if it is MM? Probably intepolate the default instance.
            ufo = builder.get_ufo_masters()[0]
            if fonttype == ftOPENTYPE:
                font = compileOTF(
                    ufo,
                    cffVersion=1,
                    inplace=True,
                    # layerName=,
                    optimizeCFF=2,
                    skipExportGlyphs=False,
                )
                font.save(filename)
                return 0
            elif fonttype == ftTRUETYPE:
                # TODO: For TTF, compile UFO hinting
                font = compileTTF(
                    ufo,
                    removeOverlaps=False,
                    flattenComponents=False,  # FL only has one level of components
                    # convertCubics=,
                    # cubicConversionError=,
                    # layerName=,
                    skipExportGlyphs=False,
                    dropImpliedOnCurves=False,
                    allQuadratic=True,
                )
                font.save(filename)
                return 0
            elif fonttype == ftFONTLAB:
                # What's the difference to just saving the VFB? Probably interpolating
                # the default instance?
                # TODO: Verify and interpolate
                font.Save(filename)
                return 0

        else:
            # We cannot generate a font at the moment, let's fake it.
            # Check if the current font has a fake_binary.
            binary = font.fake_binary_get(fonttype)
            if binary:
                with open(filename, "wb") as f:
                    f.write(binary)
                return 0

        return -1

    def Add(self, font: Font) -> None:
        """
        Add 'font' to list of opened fonts and open the Font Window for it.
        """
        self._fonts.append(font)
        self.ifont = len(self._fonts) - 1
        # As long as no glyph window is open,
        # delta and scale will be points at (0, 0)
        self.delta = Point()
        self.scale = Point()
        self._layer = 0

    def UpdateFont(self, fontindex: int | None = None) -> None:
        """
        Updates current font or 'fontindex' (slow operation)
        """
        if fontindex is None:
            fontindex = self.ifont
        self._fonts[fontindex].fake_update()

    def SetFontWindow(
        self, fontindex: int, position: Rect, state: int | None = None
    ) -> None:
        """
        Set the window size and style for the Font Window where font 'fontindex' is
        shown.

        Args:
            fontindex (int): The font index.
            position (Rect): The position and size of the window.
            state (int | None, optional): The window state.
                0 = normal, 1 = minimized, 2 = maximized. Defaults to None.
        """
        raise NotImplementedError

    def UpdateGlyph(self, glyphindex: int | None = None) -> None:
        """
        Update the current glyph or 'glyphindex' glyph of the current font
        """
        raise NotImplementedError

    def EditGlyph(self, glyphindex: int | None = None) -> None:
        """
        Open the Glyph window for the 'glyphindex' glyph in the current font
        """
        raise NotImplementedError

    def CallCommand(self, commandcode: int) -> None:
        """
        Simulate a menu or toolbar command. Check WS_* constants for list of available
        commands.
        """
        raise NotImplementedError

    def Selected(self, glyphindex: int | None = None) -> int:
        """
        Return 1 if current glyph or 'glyphindex' glyph is selected.
        """
        if self.font is None:
            return 0

        if glyphindex is None:
            raise NotImplementedError
            # return 1 if fl.glyph.selected

        if glyphindex in self.font._selection:
            return 1

        return 0

    def Select(self, glyphid: str | Uni | int, value: bool | None = None) -> None:
        """
        Change a glyph's selection state. 'glyphid' may be string (glyph name), Uni
        (Unicode index) or integer (glyph index)

        Args:
            glyphid (str | Uni | int): The glyph identifier.
            value (bool | None, optional): The selection state. Defaults to None. None
                is interpreted as True.

        Raises:
            RuntimeError: If no font is opened.
        """
        if self.font is None:
            raise RuntimeError(
                "No font is available to perform operation: FontLab.Select()"
            )
        if isinstance(glyphid, str):
            glyphid = self.font.FindGlyph(glyphid)
        if value is None:
            value = True
        self.font.fake_select(glyphid, value)

    def Unselect(self) -> None:
        """
        Deselect all glyphs in the current font (fast operation).
        """
        if self.font is None:
            raise RuntimeError(
                "No font is available to perform operation: FontLab.Unselect()"
            )
        self.font.fake_deselect_all()

    def Message(
        self,
        message: str,
        question: str | None = None,
        okstring: str | None = None,
        cancelstring: str | None = None,
    ) -> int:
        """
        Show an alert message dialog box. All parameters but the first can be omitted.

        Args:
            message (str): The message.
            question (str | None, optional): The question text. Defaults to None.
            okstring (str | None, optional): The label of the OK button. Defaults to
                None.
            cancelstring (str | None, optional): The label of the Cancel button.
                Defaults to None.

        Returns:
            int: 1 if the OK button was clicked, 2 if the Cancel button was clicked or
                the window has been closed.
        """
        raise NotImplementedError

    def ScreenToGlyph(self, position: Point) -> Point:
        """
        Converts the screen coordinates of the point to glyph coordinates in the current
        Glyph Window.
        """
        raise NotImplementedError

    def GlyphToScreen(self, position: Point) -> Point:
        """
        Converts the glyph coordinates of the point to screen coordinates in the current
        Glyph Window.
        """
        raise NotImplementedError

    def UpdateRect(self, r: Rect) -> None:
        """
        Updates the rectangle r in the current Glyph Window.
        """
        raise NotImplementedError

    def HitContour(self, p: Point) -> tuple[int, int, float] | None:
        """
        Contour hit detection in the current Glyph Window.

        Args:
            p (Point): The point to use for hit detection.

        Returns:
            tuple[int, int, float] | None : (nodeindex, nodesubindex, hit_time), or None
                if the point doesn't hit any contour.

        Raises:
            RuntimeError: If the current window is not a Glyph Window.
            RuntimeError: If no window is open.
        """
        # RuntimeError: No window is opened: FontLab.HitContour(Point)
        # RuntimeError: Current window is not Glyph window: FontLab.HitContour(Point)
        raise NotImplementedError

    def GetCanvas(self) -> Canvas:
        """
        Return a :py:class:`Canvas` for the current Glyph Window.
        """
        raise RuntimeError("Current window is not Glyph window: FontLab.GetCanvas()")

    def GetConvert(self, c: Canvas) -> None:
        """
        Copy conversion parameters from the current Glyph Window to the `Canvas`.
        """
        raise NotImplementedError

    def BeginProgress(self, title: str, counts: int) -> None:
        """
        Open a Progress dialog box with 'counts' number of 'ticks'.
        """
        raise NotImplementedError

    def TickProgress(self, tick: int) -> bool:
        """
        Update the Progress bar, return False if the Cancel button was pressed.
        This is a relatively 'expensive' operation.
        """
        raise NotImplementedError

    def EndProgress(self) -> None:
        """
        Close the Progress dialog box.
        """
        raise NotImplementedError

    def Random(self, lovalue: float, hivalue: float | None = None) -> int:
        """
        (hivalue) | (lovalue, hivalue)

        - returns random value (fast operation)
        """
        # lovalue and hivalue are truncated to int before finding the random number
        # int(lovalue) <= n < int(hivalue)
        raise NotImplementedError

    def TransformGlyph(self, glyph: Glyph, code: int, text: str) -> None:
        """
        Transforms the glyph using one of the Transform actions.

        Save one of Transform Range programs for reference to format of the command
        string.
        """
        raise NotImplementedError

    def ForSelected(self, function_name: str) -> None:
        """
        Call 'function_name' for each selected glyph in the current font. The function
        has the following format:

        function(font: Font, glyph: Glyph, glyphindex: int)
        """
        raise NotImplementedError

    def SetUndo(self) -> None:
        """
        Save the current state to the undo buffer.
        """
        raise NotImplementedError

    # Additional methods reported by dir(fl)

    def GetFileName(
        self,
        open_save: int,
        suffix: str | None = None,
        filename: str | None = None,
        filetype: str | None = None,
    ) -> str:
        """
        Open a file picker dialog.

        Args:
            open_save (int): 0 for a Save dialog, 1 for an Open dialog.
            suffix (str | None, optional): The suggested file suffix. Ignored in Open
                dialogs. Defaults to None.
            filename (str | None, optional): The suggested file name. Defaults to None.
            filetype (str | None, optional): The suggested file type. Defaults to None.

        Returns:
            str: The path of the selected file.
        """
        raise NotImplementedError

    def GetPathName(self, title: str, message: str) -> str:
        """
        Show a folder picker.

        Args:
            title (str): The title of the dialog.
            message (str): The message of the dialog

        Returns:
            str: The path of the selected folder.
        """
        raise NotImplementedError

    def LoadNamFile(self, namfilename: str) -> None:
        raise NotImplementedError

    def Lock(self, lock: bool) -> None:
        raise NotImplementedError

    def OpenFont(self, filename: str) -> Font | None:
        """
        Open a VFB file, or import a font from path `filename`. If successful, the
        `Font` is returned. The font is not added as a visible Window.

        Args:
            filename (str): The path to the VFB or font.

        Returns:
            Font | None: The font, or None if the file could not be opened.
        """
        font = Font()
        try:
            font.Open(filename)
        except:  # noqa: E722
            return None
        return font

    def _setcurrentfon(self) -> None:
        # What is this for?
        raise NotImplementedError



from __future__ import annotations

from FL.fake.Base import Copyable


class EncodingRecord(Copyable):
    __slots__ = ["_name", "_unicode"]

    # Constructor

    def __init__(self, encoding_record: EncodingRecord | None = None) -> None:
        """
        EncodingRecord - class to represent EncodingRecord

        EncodingRecord()
            generic constructor, creates empty EncodingRecord
        EncodingRecord(EncodingRecord)
            copy constructor

        Args:
            encoding_record (EncodingRecord | None, optional): _description_. Defaults to None.
        """
        self.name = "(null)"
        self.unicode = 0
        if isinstance(encoding_record, EncodingRecord):
            self._copy_constructor(encoding_record)

    def __repr__(self) -> str:
        return f'<EncodingRecord: "{self.name}", unicode: {self.unicode}>'

    # Attributes

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def unicode(self) -> int:
        return self._unicode

    @unicode.setter
    def unicode(self, value: int) -> None:
        self._unicode = value

    # Operations

    # EncodingRecord has no operations

    # Methods

    # EncodingRecord has no methods



from __future__ import annotations

from vfbLib.compilers.truetype import convert_flags_options_to_int
from vfbLib.helpers import deHexStr, hexStr
from vfbLib.parsers.truetype import convert_int_to_flags_options
from vfbLib.typing import TTStemsDict, TTZonesDict

from FL.fake.Base import Copyable
from FL.helpers.math import int32_to_unsigned, uint32_to_signed
from FL.objects.TTGasp import TTGasp
from FL.objects.TTStem import TTStem
from FL.objects.TTVdmx import TTVdmx


class TTInfo(Copyable):
    """
    TTInfo - class to represent TrueType Information

    All information about TTInfo is based on an eMail from Yuri and was valid for
    FontLab 4.0
    """

    __slots__ = [
        "_hstem_data",
        "_vstem_data",
        "_prep",
        "_fpgm",
        "_cvt",
        "_gasp",
        "_hdmx",
        "_vdmx",
        "_max_zones",
        "_max_twilight_points",
        "_max_storage",
        "_max_function_defs",
        "_max_instruction_defs",
        "_max_stack_elements",
        "_head_creation",
        "_head_flags",
        "_head_font_direction_hint",
        "_head_lowest_rec_ppem",
        "_head_mac_style",
        "_head_units_per_em",
        "_hhea_ascender",
        "_hhea_descender",
        "_hhea_line_gap",
        "_os2_us_weight_class",
        "_os2_us_width_class",
        "_os2_fs_type",
        "_os2_y_subscript_x_size",
        "_os2_y_subscript_y_size",
        "_os2_y_subscript_x_offset",
        "_os2_y_subscript_y_offset",
        "_os2_y_superscript_x_size",
        "_os2_y_superscript_y_size",
        "_os2_y_superscript_x_offset",
        "_os2_y_superscript_y_offset",
        "_os2_y_strikeout_size",
        "_os2_y_strikeout_position",
        "_os2_s_family_class",
        "_os2_ul_code_page_range1",
        "_os2_ul_code_page_range2",
        "_os2_s_typo_ascender",
        "_os2_s_typo_descender",
        "_os2_s_typo_line_gap",
        "_os2_fs_selection",
        "_os2_us_win_ascent",
        "_os2_us_win_descent",
        # Non-API additions:
        "_average_width",
        "_codepages",
        "_hdmx_ppms_1",
        "_hdmx_ppms_2",
        "_panose",
        "_stemsnaplimit",
        "_zoneppm",
        "_codeppm",
        "_zones",
        "_unknown_pleasures",
    ]

    # Constructor

    def __init__(self, ttinfo: TTInfo | None = None) -> None:
        """
        Never create TTInfo object explicitly - they must be obtained from the FontLab's
        `Font` classes.

        Args:
            ttinfo (TTInfo | None): Copy attributes from `TTInfo` or create an empty
                `TTInfo`
        """
        if ttinfo is None:
            self._set_defaults()
        else:
            self._copy_constructor(ttinfo)

    # Additions for FakeLab

    def fake_deserialize(
        self, data: dict[str, int | list[int] | dict[str, int]]
    ) -> None:
        for k in (
            "max_zones",
            "max_twilight_points",
            "max_storage",
            "max_function_defs",
            "max_stack_elements",
            # "head_flags",  # Different format in JSON
            "head_units_per_em",
            "head_mac_style",
            "head_lowest_rec_ppem",
            "head_font_direction_hint",
            "os2_us_weight_class",
            "os2_us_width_class",
            "os2_fs_type",
            "os2_y_subscript_x_size",
            "os2_y_subscript_y_size",
            "os2_y_subscript_x_offset",
            "os2_y_subscript_y_offset",
            "os2_y_superscript_x_size",
            "os2_y_superscript_y_size",
            "os2_y_superscript_x_offset",
            "os2_y_superscript_y_offset",
            "os2_y_strikeout_size",
            "os2_y_strikeout_position",
            "os2_s_family_class",
            "os2_s_typo_ascender",
            "os2_s_typo_descender",
            "os2_s_typo_line_gap",
            "os2_fs_selection",
            "os2_us_win_ascent",
            "os2_us_win_descent",
        ):
            if k in data:
                setattr(self, k, data[k])
        for k, attr in (
            ("Average Width", "_average_width"),
            ("OpenTypeOS2Panose", "_panose"),
            ("Hdmx PPMs 1", "_hdmx_ppms_1"),
            ("Hdmx PPMs 2", "_hdmx_ppms_2"),
            ("Codepages", "_codepages"),
        ):
            if k in data:
                setattr(self, attr, data[k])
        if "head_creation" in data:
            self._head_creation[0] = uint32_to_signed(data["head_creation"])
        if "head_creation2" in data:
            self._head_creation[1] = uint32_to_signed(data["head_creation2"])

        # I tried to make this value more self-explanatory in vfbLib, but FL
        # only shows us an int. So we have to convert it back ...
        if "head_flags" in data:
            self.head_flags = convert_flags_options_to_int(data)

    def fake_serialize(self) -> dict[str | int, int | list[int] | dict[str, int]]:
        return {
            "max_zones": self.max_zones,
            "max_twilight_points": self.max_twilight_points,
            "max_storage": self.max_storage,
            "max_function_defs": self.max_function_defs,
            "max_instruction_defs": self.max_instruction_defs,
            "max_stack_elements": self.max_stack_elements,
            "head_flags": convert_int_to_flags_options(self.head_flags),
            "head_units_per_em": self.head_units_per_em,
            "head_mac_style": self.head_mac_style,
            "head_lowest_rec_ppem": self.head_lowest_rec_ppem,
            "head_creation": int32_to_unsigned(self.head_creation[0]),
            "head_creation2": int32_to_unsigned(self.head_creation[1]),
            "head_font_direction_hint": self.head_font_direction_hint,
            "os2_us_weight_class": self.os2_us_weight_class,
            "os2_us_width_class": self.os2_us_width_class,
            "os2_fs_type": self.os2_fs_type,
            "os2_y_subscript_x_size": self.os2_y_subscript_x_size,
            "os2_y_subscript_y_size": self.os2_y_subscript_y_size,
            "os2_y_subscript_x_offset": self.os2_y_subscript_x_offset,
            "os2_y_subscript_y_offset": self.os2_y_subscript_y_offset,
            "os2_y_superscript_x_size": self.os2_y_superscript_x_size,
            "os2_y_superscript_y_size": self.os2_y_superscript_y_size,
            "os2_y_superscript_x_offset": self.os2_y_superscript_x_offset,
            "os2_y_superscript_y_offset": self.os2_y_superscript_y_offset,
            "os2_y_strikeout_size": self.os2_y_strikeout_size,
            "os2_y_strikeout_position": self.os2_y_strikeout_position,
            "os2_s_family_class": self.os2_s_family_class,
            "OpenTypeOS2Panose": self._panose,
            "os2_s_typo_ascender": self.os2_s_typo_ascender,
            "os2_s_typo_descender": self.os2_s_typo_descender,
            "os2_s_typo_line_gap": self.os2_s_typo_line_gap,
            "os2_fs_selection": self.os2_fs_selection,
            "os2_us_win_ascent": self.os2_us_win_ascent,
            "os2_us_win_descent": self.os2_us_win_descent,
            "Average Width": self._average_width,
            "Hdmx PPMs 1": self._hdmx_ppms_1,
            "Hdmx PPMs 2": self._hdmx_ppms_2,
            "Codepages": self._codepages,
        }

    def fake_deserialize_gasp(self, data: list[dict[str, int]]) -> None:
        for rec in data:
            gasp_rec = TTGasp()
            gasp_rec.fake_deserialize(rec)
            self.gasp.append(gasp_rec)

    def fake_serialize_gasp(self) -> list[dict[str, int]]:
        gasp = []
        for rec in self.gasp:
            gasp.append(rec.fake_serialize())
        return gasp

    def fake_deserialize_stems(self, data: TTStemsDict) -> None:
        # Some TrueType stem entries are optional, so the stems lists may have been
        # populated yet
        if not (self.hstem_data or self.vstem_data):
            # Stem lists are empty, make new stems
            for stem_dict in data["ttStemsV"]:
                ttstem = TTStem()
                ttstem.fake_deserialize(stem_dict)
                self.hstem_data.append(ttstem)
            for stem_dict in data["ttStemsH"]:
                ttstem = TTStem()
                ttstem.fake_deserialize(stem_dict)
                self.vstem_data.append(ttstem)
        else:
            for i, stem_dict in enumerate(data["ttStemsV"]):
                ttstem = self.hstem_data[i]
                ttstem.fake_deserialize(stem_dict)
            for i, stem_dict in enumerate(data["ttStemsH"]):
                ttstem = self.vstem_data[i]
                ttstem.fake_deserialize(stem_dict)

    def fake_serialize_stem_ppems(self) -> TTStemsDict:
        d: TTStemsDict = {"ttStemsV": [], "ttStemsH": []}
        for i, ttstem in enumerate(self.hstem_data):
            d["ttStemsV"].append(
                {
                    "stem": i,
                    "round": {
                        "2": ttstem.ppm2,
                        "3": ttstem.ppm3,
                        "4": ttstem.ppm4,
                        "5": ttstem.ppm5,
                    },
                }
            )
        for i, ttstem in enumerate(self.vstem_data):
            d["ttStemsH"].append(
                {
                    "stem": i,
                    "round": {
                        "2": ttstem.ppm2,
                        "3": ttstem.ppm3,
                        "4": ttstem.ppm4,
                        "5": ttstem.ppm5,
                    },
                }
            )
        return d

    def fake_serialize_stem_ppems1(self) -> TTStemsDict:
        d: TTStemsDict = {"ttStemsV": [], "ttStemsH": []}
        for i, ttstem in enumerate(self.hstem_data):
            d["ttStemsV"].append({"stem": i, "round": {"1": ttstem._ppm1}})
        for i, ttstem in enumerate(self.vstem_data):
            d["ttStemsH"].append({"stem": i, "round": {"1": ttstem._ppm1}})
        return d

    def fake_serialize_stems(self) -> TTStemsDict:
        d: TTStemsDict = {"ttStemsV": [], "ttStemsH": []}
        for ttstem in self.hstem_data:
            d["ttStemsV"].append(
                {
                    "value": ttstem.width,
                    "name": ttstem.name,
                    "round": {
                        "6": ttstem.ppm6,
                    },
                }
            )
        for ttstem in self.vstem_data:
            d["ttStemsH"].append(
                {
                    "value": ttstem.width,
                    "name": ttstem.name,
                    "round": {
                        "6": ttstem.ppm6,
                    },
                }
            )
        return d

    def fake_deserialize_vdmx(self, data: list[dict[str, int]]) -> None:
        for rec in data:
            vdmx_rec = TTVdmx()
            vdmx_rec.fake_deserialize(rec)
            self.vdmx.append(vdmx_rec)

    def fake_serialize_vdmx(self) -> list[dict[str, int]]:
        vdmx = []
        for rec in self.vdmx:
            vdmx.append(rec.fake_serialize())
        return vdmx

    def fake_deserialize_zones(self, data: TTZonesDict) -> None:
        self._zones = data

    def fake_serialize_zones(self) -> TTZonesDict:
        d: TTZonesDict = {"ttZonesT": [], "ttZonesB": []}
        for side in ("ttZonesT", "ttZonesB"):
            for zone in self._zones[side]:
                d[side].append(
                    {
                        "position": zone["position"],
                        "value": zone["value"],
                        "name": zone["name"],
                    }
                )
        return d

    def fake_deserialize_zone_deltas(self, data: dict[int, dict[int, int]]) -> None:
        num_bottom_zones = len(self._zones["ttZonesB"])
        for zone_index, deltas in data.items():
            if zone_index < num_bottom_zones:
                zone = self._zones["ttZonesB"][zone_index]
            else:
                zone = self._zones["ttZonesT"][zone_index - num_bottom_zones]
            zone["deltas"] = deltas

    def fake_serialize_zone_deltas(self) -> dict[int, dict[int, int]]:
        d: dict[int, dict[int, int]] = {}
        # Zone deltas are indexed, bottom zones first
        bottom_index = 0
        for bottom_index, zone_dict in enumerate(self._zones["ttZonesB"]):
            zone_deltas = zone_dict.get("deltas", {})
            if zone_deltas:
                d[bottom_index] = zone_deltas
        for top_index, zone_dict in enumerate(self._zones["ttZonesT"]):
            zone_deltas = zone_dict.get("deltas", {})
            if zone_deltas:
                d[bottom_index + top_index + 1] = zone_deltas
        return d

    def fake_get_binary(self, attr: str) -> str:
        return hexStr(bytes(getattr(self, attr)))

    def fake_set_binary(self, attr: str, data: str) -> None:
        """
        Set a binary table (cvt, prep, fpgm) from the representation chosen in JSON.

        Args:
            attr (str): The attribute name the data should go to
            data (str): The data as a hexStr
        """
        setattr(self, attr, list(deHexStr(data)))

    @property
    def hstem_data(self) -> list[TTStem]:
        """
        list of horizontal TrueType-Stems 'VExTTStemArray'

        Returns:
            list[TTStem]: _description_
        """
        return self._hstem_data

    @hstem_data.setter
    def hstem_data(self, value: list[TTStem]) -> None:
        self._hstem_data = value

    @property
    def vstem_data(self) -> list[TTStem]:
        """
        list of vertical TrueType-Stems 'VExTTStemArray'

        Returns:
            list[TTStem]: _description_
        """
        return self._vstem_data

    @vstem_data.setter
    def vstem_data(self, value: list[TTStem]) -> None:
        self._vstem_data = value

    @property
    def prep(self) -> list[int]:
        """
        CVT Program: list 'VSICharArray'

        Returns:
            list[int]: _description_
        """
        return self._prep

    @prep.setter
    def prep(self, value: list[int]) -> None:
        self._prep = value

    @property
    def fpgm(self) -> list[int]:
        """
        Font program: list 'VSICharArray'

        Returns:
            list[int]: _description_
        """
        return self._fpgm

    @fpgm.setter
    def fpgm(self, value: list[int]) -> None:
        self._fpgm = value

    @property
    def cvt(self) -> list[int]:
        """
        Control Value Table: list 'VSICharArray'

        Returns:
            list[int]: _description_
        """
        return self._cvt

    @cvt.setter
    def cvt(self, value: list[int]) -> None:
        self._cvt = value

    @property
    def gasp(self) -> list[TTGasp]:
        """
        Grid-fitting/Scan-conversion: list of smoothing control records 'VSArray'

        Returns:
            list[TTGasp]: _description_
        """
        return self._gasp

    @gasp.setter
    def gasp(self, value: list[TTGasp]) -> None:
        self._gasp = value

    @property
    def hdmx(self) -> list[int]:
        """
        Horizontal device metrics: list 'VSBByteArray'

        Returns:
            list[int]: _description_
        """
        return self._hdmx

    @hdmx.setter
    def hdmx(self, value: list[int]) -> None:
        self._hdmx = value

    @property
    def vdmx(self) -> list[TTVdmx]:
        """
        Vertical device metrics: list 'VSArray'

        Returns:
            list[TTVdmx]: _description_
        """
        return self._vdmx

    @vdmx.setter
    def vdmx(self, value: list[TTVdmx]) -> None:
        self._vdmx = value

    @property
    def max_zones(self) -> int:
        return self._max_zones

    @max_zones.setter
    def max_zones(self, value: int) -> None:
        self._max_zones = value

    @property
    def max_twilight_points(self) -> int:
        return self._max_twilight_points

    @max_twilight_points.setter
    def max_twilight_points(self, value: int) -> None:
        self._max_twilight_points = value

    @property
    def max_storage(self) -> int:
        return self._max_storage

    @max_storage.setter
    def max_storage(self, value: int) -> None:
        self._max_storage = value

    @property
    def max_function_defs(self) -> int:
        return self._max_function_defs

    @max_function_defs.setter
    def max_function_defs(self, value: int) -> None:
        self._max_function_defs = value

    @property
    def max_instruction_defs(self) -> int:
        return self._max_instruction_defs

    @max_instruction_defs.setter
    def max_instruction_defs(self, value: int) -> None:
        self._max_instruction_defs = value

    @property
    def max_stack_elements(self) -> int:
        return self._max_stack_elements

    @max_stack_elements.setter
    def max_stack_elements(self, value: int) -> None:
        self._max_stack_elements = value

    @property
    def head_creation(self) -> list[int]:
        return self._head_creation

    @head_creation.setter
    def head_creation(self, value: list[int]) -> None:
        raise RuntimeError(
            "class TTInfo has no attribute head_creation or it's read-only"
        )

    @property
    def head_flags(self) -> int:
        return self._head_flags

    @head_flags.setter
    def head_flags(self, value: int) -> None:
        self._head_flags = value

    @property
    def head_font_direction_hint(self) -> int:
        return self._head_font_direction_hint

    @head_font_direction_hint.setter
    def head_font_direction_hint(self, value: int) -> None:
        self._head_font_direction_hint = value

    @property
    def head_lowest_rec_ppem(self) -> int:
        return self._head_lowest_rec_ppem

    @head_lowest_rec_ppem.setter
    def head_lowest_rec_ppem(self, value: int) -> None:
        self._head_lowest_rec_ppem = value

    @property
    def head_mac_style(self) -> int:
        return self._head_mac_style

    @head_mac_style.setter
    def head_mac_style(self, value: int) -> None:
        self._head_mac_style = value

    @property
    def head_units_per_em(self) -> int:
        return self._head_units_per_em

    @head_units_per_em.setter
    def head_units_per_em(self, value: int) -> None:
        self._head_units_per_em = value

    @property
    def hhea_ascender(self) -> int:
        return self._hhea_ascender

    @hhea_ascender.setter
    def hhea_ascender(self, value: int) -> None:
        self._hhea_ascender = value

    @property
    def hhea_descender(self) -> int:
        return self._hhea_descender

    @hhea_descender.setter
    def hhea_descender(self, value: int) -> None:
        self._hhea_descender = value

    @property
    def hhea_line_gap(self) -> int:
        return self._hhea_line_gap

    @hhea_line_gap.setter
    def hhea_line_gap(self, value: int) -> None:
        self._hhea_line_gap = value

    @property
    def os2_us_weight_class(self) -> int:
        return self._os2_us_weight_class

    @os2_us_weight_class.setter
    def os2_us_weight_class(self, value: int) -> None:
        self._os2_us_weight_class = value

    @property
    def os2_us_width_class(self) -> int:
        return self._os2_us_width_class

    @os2_us_width_class.setter
    def os2_us_width_class(self, value: int) -> None:
        self._os2_us_width_class = value

    @property
    def os2_fs_type(self) -> int:
        return self._os2_fs_type

    @os2_fs_type.setter
    def os2_fs_type(self, value: int) -> None:
        self._os2_fs_type = value

    @property
    def os2_y_subscript_x_size(self) -> int:
        return self._os2_y_subscript_x_size

    @os2_y_subscript_x_size.setter
    def os2_y_subscript_x_size(self, value: int) -> None:
        self._os2_y_subscript_x_size = value

    @property
    def os2_y_subscript_y_size(self) -> int:
        return self._os2_y_subscript_y_size

    @os2_y_subscript_y_size.setter
    def os2_y_subscript_y_size(self, value: int) -> None:
        self._os2_y_subscript_y_size = value

    @property
    def os2_y_subscript_x_offset(self) -> int:
        return self._os2_y_subscript_x_offset

    @os2_y_subscript_x_offset.setter
    def os2_y_subscript_x_offset(self, value: int) -> None:
        self._os2_y_subscript_x_offset = value

    @property
    def os2_y_subscript_y_offset(self) -> int:
        return self._os2_y_subscript_y_offset

    @os2_y_subscript_y_offset.setter
    def os2_y_subscript_y_offset(self, value: int) -> None:
        self._os2_y_subscript_y_offset = value

    @property
    def os2_y_superscript_x_size(self) -> int:
        return self._os2_y_superscript_x_size

    @os2_y_superscript_x_size.setter
    def os2_y_superscript_x_size(self, value: int) -> None:
        self._os2_y_superscript_x_size = value

    @property
    def os2_y_superscript_y_size(self) -> int:
        return self._os2_y_superscript_y_size

    @os2_y_superscript_y_size.setter
    def os2_y_superscript_y_size(self, value: int) -> None:
        self._os2_y_superscript_y_size = value

    @property
    def os2_y_superscript_x_offset(self) -> int:
        return self._os2_y_superscript_x_offset

    @os2_y_superscript_x_offset.setter
    def os2_y_superscript_x_offset(self, value: int) -> None:
        self._os2_y_superscript_x_offset = value

    @property
    def os2_y_superscript_y_offset(self) -> int:
        return self._os2_y_superscript_y_offset

    @os2_y_superscript_y_offset.setter
    def os2_y_superscript_y_offset(self, value: int) -> None:
        self._os2_y_superscript_y_offset = value

    @property
    def os2_y_strikeout_size(self) -> int:
        return self._os2_y_strikeout_size

    @os2_y_strikeout_size.setter
    def os2_y_strikeout_size(self, value: int) -> None:
        self._os2_y_strikeout_size = value

    @property
    def os2_y_strikeout_position(self) -> int:
        return self._os2_y_strikeout_position

    @os2_y_strikeout_position.setter
    def os2_y_strikeout_position(self, value: int) -> None:
        self._os2_y_strikeout_position = value

    @property
    def os2_s_family_class(self) -> int:
        return self._os2_s_family_class

    @os2_s_family_class.setter
    def os2_s_family_class(self, value: int) -> None:
        self._os2_s_family_class = value

    @property
    def os2_ul_code_page_range1(self) -> int:
        return self._os2_ul_code_page_range1

    @os2_ul_code_page_range1.setter
    def os2_ul_code_page_range1(self, value: int) -> None:
        self._os2_ul_code_page_range1 = value

    @property
    def os2_ul_code_page_range2(self) -> int:
        return self._os2_ul_code_page_range2

    @os2_ul_code_page_range2.setter
    def os2_ul_code_page_range2(self, value: int) -> None:
        self._os2_ul_code_page_range2 = value

    @property
    def os2_s_typo_ascender(self) -> int:
        return self._os2_s_typo_ascender

    @os2_s_typo_ascender.setter
    def os2_s_typo_ascender(self, value: int) -> None:
        self._os2_s_typo_ascender = value

    @property
    def os2_s_typo_descender(self) -> int:
        return self._os2_s_typo_descender

    @os2_s_typo_descender.setter
    def os2_s_typo_descender(self, value: int) -> None:
        self._os2_s_typo_descender = value

    @property
    def os2_s_typo_line_gap(self) -> int:
        return self._os2_s_typo_line_gap

    @os2_s_typo_line_gap.setter
    def os2_s_typo_line_gap(self, value: int) -> None:
        self._os2_s_typo_line_gap = value

    @property
    def os2_fs_selection(self) -> int:
        return self._os2_fs_selection

    @os2_fs_selection.setter
    def os2_fs_selection(self, value: int) -> None:
        self._os2_fs_selection = value

    @property
    def os2_us_win_ascent(self) -> int:
        return self._os2_us_win_ascent

    @os2_us_win_ascent.setter
    def os2_us_win_ascent(self, value: int) -> None:
        self._os2_us_win_ascent = value

    @property
    def os2_us_win_descent(self) -> int:
        return self._os2_us_win_descent

    @os2_us_win_descent.setter
    def os2_us_win_descent(self, value: int) -> None:
        self._os2_us_win_descent = value

    # Internal

    def _set_defaults(self) -> None:
        # Defaults for an empty TTInfo
        self.hstem_data = []
        self.vstem_data = []
        self.prep = []
        self.fpgm = []
        self.cvt = []
        self.gasp = []
        self.vdmx = []
        self.max_zones = 0
        self.max_twilight_points = 0
        self.max_storage = 0
        self.max_function_defs = 0
        self.max_instruction_defs = 0
        self.max_stack_elements = 0
        self._head_creation: list[int] = [-467938523, 0]
        self.head_flags = 131072
        self.head_font_direction_hint = 2
        self.head_lowest_rec_ppem = 9
        self.head_mac_style = 0
        self.head_units_per_em = 2048
        self.hhea_ascender = 1536
        self.hhea_descender = -512
        self.hhea_line_gap = 18
        self.os2_us_weight_class = 400
        self.os2_us_width_class = 5
        self.os2_fs_type = 4
        self.os2_y_subscript_x_size = 1433
        self.os2_y_subscript_y_size = 1331
        self.os2_y_subscript_x_offset = 0
        self.os2_y_subscript_y_offset = 286
        self.os2_y_superscript_x_size = 1433
        self.os2_y_superscript_y_size = 1331
        self.os2_y_superscript_x_offset = 0
        self.os2_y_superscript_y_offset = 976
        self.os2_y_strikeout_size = 102
        self.os2_y_strikeout_position = 512
        self.os2_s_family_class = 0
        self.os2_ul_code_page_range1 = 0
        self.os2_ul_code_page_range2 = 0
        self.os2_s_typo_ascender = 1536
        self.os2_s_typo_descender = -512
        self.os2_s_typo_line_gap = 0
        self.os2_fs_selection = 0
        self.os2_us_win_ascent = 0
        self.os2_us_win_descent = 0

        # Not in API:
        self._average_width = 0
        self._codepages: dict[str, int] = {
            "os2_ul_code_page_range1": 0,
            "os2_ul_code_page_range2": 0,
        }
        self._hdmx_ppms_1: list[int] = []
        self._hdmx_ppms_2: list[int] = []
        self._panose: list[int] = [0] * 10
        # TT-related non-API:
        self._stemsnaplimit: int = 68  # 68/64 pixel
        self._zoneppm: int = 48  # Zones active until ppm
        self._codeppm: int = 0  # Gridfitting active until ppm (0 = no limit)

        self._zones: TTZonesDict = {"ttZonesT": [], "ttZonesB": []}

        self._unknown_pleasures = {
            1604: 255,
            2032: 0,
        }



import unittest

from vfbLib.enum import G

from FL.objects.Glyph import Glyph
from FL.objects.Point import Point

glyph_a = {
    "name": "a",
    "num_masters": 1,
    "nodes": [
        {"type": "move", "flags": 3, "points": [[(77, 359)]]},
        {"type": "curve", "flags": 3, "points": [[(272, 106), (77, 219), (164, 106)]]},
        {"type": "curve", "flags": 3, "points": [[(468, 359), (380, 106), (468, 219)]]},
        {"type": "curve", "flags": 3, "points": [[(272, 613), (468, 499), (380, 613)]]},
        {"type": "curve", "flags": 3, "points": [[(77, 359), (164, 613), (77, 499)]]},
    ],
    "metrics": [[556, 0]],
    "kerning": {4: [-100]},
    "components": [],
}


class GlyphTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        g = Glyph()
        assert g.parent is None

    def test_deserialize(self) -> None:
        # Deserialize the Glyph object from json.
        g = Glyph()
        g.fake_deserialize(G.Glyph, glyph_a)
        assert g.name == "a"
        assert len(g) == 5  # number of nodes
        assert g.GetMetrics() == Point(556, 0)
        assert len(g.kerning) == 1

        s = g.fake_serialize()
        assert s[G.Glyph] == glyph_a



from __future__ import annotations

from FL.fake.Base import Copyable


class TTVdmx(Copyable):
    """
    A vdmx table record.

    This class is completely undocumented.
    """

    __slots__ = ["_y_pel_height", "_y_min", "_y_max"]

    def __init__(self, ttvdmx: TTVdmx | None = None) -> None:
        self._y_pel_height = 0
        self._y_min = 0
        self._y_max = 0
        if isinstance(ttvdmx, TTVdmx):
            self._copy_constructor(ttvdmx)

    def __repr__(self) -> str:
        return (
            f"<TTVdmx: PelHeight: {self.y_pel_height}, Min: {self.y_min}, Max: "
            f"{self.y_max}, Orphan>"
        )

    # Additions for FakeLab

    def fake_deserialize(self, data: dict[str, int]) -> None:
        self.y_pel_height = data["pelHeight"]
        self.y_max = data["max"]
        self.y_min = data["min"]

    def fake_serialize(self) -> dict[str, int]:
        return {"pelHeight": self.y_pel_height, "max": self.y_max, "min": self.y_min}

    # Attributes

    @property
    def y_pel_height(self) -> int:
        return self._y_pel_height

    @y_pel_height.setter
    def y_pel_height(self, value: int) -> None:
        self._y_pel_height = value

    @property
    def y_min(self) -> int:
        return self._y_min

    @y_min.setter
    def y_min(self, value: int) -> None:
        self._y_min = value

    @property
    def y_max(self) -> int:
        return self._y_max

    @y_max.setter
    def y_max(self, value: int) -> None:
        self._y_max = value



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Font import Font


class WeightVector(Copyable):
    """This class is completely undocumented.

    Apparently: A list of 16 floats that represent the interpolation factors for all
    possible masters.
    """

    __slots__ = ["_parent", "_weights"]

    # Constructor

    def __init__(self, weight_vector: WeightVector | None = None) -> None:
        self._parent: Font | None = None
        self._weights = [
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
        if weight_vector is not None:
            self._copy_constructor(weight_vector)

    # Attributes

    @property
    def parent(self) -> Font | None:
        return self._parent

    def __getitem__(self, index: int) -> float:
        return self._weights[index]



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Font import Font


class Feature(Copyable):  # Or TagObject?
    """
    OpenType feature-definition record
    """

    __slots__ = ["_parent", "_tag", "_value"]

    # Constructor

    def __init__(
        self, feature_or_tag: Feature | str | None = None, value: str | None = None
    ) -> None:
        """
        - `Feature()`: generic constructor, creates an empty Feature record
        - `Feature(Feature)`: copy constructor
        - `Feature(string tag)`: creates feature, assigns 'tag' and empty value
        - `Feature(string tag, string value)`: creates feature and assigns values to both
          attributes

        Args:
            feature_or_tag (Feature | str | None, optional): _description_. Defaults to None.
            value (str | None, optional): _description_. Defaults to None.
        """
        self._parent: Font | None = None
        self._tag = ""
        self._value: str | None = None

        # Process params

        if isinstance(feature_or_tag, Feature):
            self._copy_constructor(feature_or_tag)

        elif isinstance(feature_or_tag, str):
            self.tag = feature_or_tag
            if value is not None:
                self.value = value
        # else: Empty Feature

    @property
    def parent(self) -> Font | None:
        """
        Returns:
            Font | None: The parent Font object, or None.
        """
        return self._parent

    @property
    def tag(self) -> str:
        """
        Returns:
            str: The four-character feature tag.
        """
        return self._tag

    @tag.setter
    def tag(self, value: str) -> None:
        self._tag = value

    @property
    def value(self) -> str | None:
        """
        Returns:
            str | None: The feature code in AFDKO syntax.
        """
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("CXX: type error.")
        self._value = value



import unittest

import pytest

from FL.objects.Hint import Hint


class HintTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        h = Hint()
        assert h.parent is None
        assert h.position == 0
        assert h.width == 21
        assert h.positions == [0] * 16
        assert h.widths == [21] * 16

    def test_instantiation_copy_extra(self) -> None:
        h0 = Hint(-1, 100)
        h = Hint(h0)
        assert h.parent is None
        assert h.position == -1
        assert h.width == 100
        assert h.positions == [-1] * 16
        assert h.widths == [100] * 16

    def test_instantiation_copy(self) -> None:
        h0 = Hint(-1, 100)
        with pytest.raises(TypeError):
            # When the second arg is an int, the first one must be as well
            Hint(h0, 1)

    def test_position_affects_mm(self) -> None:
        h = Hint()
        h.position = 50
        assert h.position == 50
        assert h.positions == [50] * 16

    def test_width_affects_mm(self) -> None:
        h = Hint()
        h.width = 50
        assert h.width == 50
        assert h.widths == [50] * 16



import unittest

from FL.objects.Point import Point
from FL.objects.Rect import Rect


class RectTests(unittest.TestCase):
    def test_instantiation(self):
        # No args
        r = Rect()
        assert r.x == 0
        assert r.y == 0
        assert r.width == 0
        assert r.height == 0
        assert r.ll == Point(0, 0)
        assert r.ur == Point(0, 0)

    def test_instantiation_point(self):
        # Documentation for Rect from Point is wrong: It says the rect will be from
        # (0, 0) to the point, but it is at the point with width = 0 and height = 0.
        r = Rect(Point(100, 100))
        assert r.x == 100
        assert r.y == 100
        assert r.width == 0
        assert r.height == 0

    def test_instantiation_coords(self):
        # Floats are preserved
        r = Rect(0.0, 0.0, 10.5, 10.5)
        assert r.x == 0
        assert r.y == 0
        assert r.width == 10.5
        assert r.height == 10.5

    def test_instantiation_normalization(self):
        # Points are not normalized
        r = Rect(10.5, 10.5, 0.0, 0.0)
        assert r.x == 10.5
        assert r.y == 10.5
        assert r.width == -10.5
        assert r.height == -10.5

    def test_various(self):
        # Directly from FL5, wtf is happening here?
        # FIXME
        r = Rect(1, 2, 3, 4)
        r.x = 5
        assert r.x == 5
        assert r.width == -2
        r.width = 100
        assert r.x == 105
        assert r.ll == Point(105, 2)
        assert r.width == -102

    # Addition

    def test_add(self):
        r = Rect(0.0, 0.0, 10.5, 10.5)
        r2 = r + Point(-1, -1)
        assert r.ll == Point(-1, -1)
        assert r.ur == Point(10.5, 10.5)
        assert r2.width == 11.5
        assert r2.height == 11.5

        # No normalization again:
        r = Rect(10.5, 10.5, 0.0, 0.0)
        r2 = r + Point(-1, -1)
        assert r.ll == Point(-1, -1)
        assert r.ur == Point(0, 0)

    def test_iadd_point(self):
        r = Rect(0.0, 0.0, 10.5, 10.5)
        r += Point(-1, -1)
        assert r.x == -1
        assert r.y == -1
        assert r.width == 11.5
        assert r.height == 11.5

        r = Rect(Point(0, -200))
        assert r.x == 0
        assert r.y == -200
        assert r.width == 0
        assert r.height == 0
        r += Point(500, -200)
        assert r.x == 0
        assert r.y == -200
        assert r.width == 500
        assert r.height == 0
        r += Point(490, -102)
        assert r.ll == Point(0, -200)
        assert r.ur == Point(500, -102)
        r += Point(-10, -102)
        assert r.ll == Point(-10, -200)
        assert r.ur == Point(500, -102)
        assert r.width == 510
        assert r.height == 98

    def test_Validate(self):
        r = Rect(105, 2, 3, 4)
        r.Validate()
        assert r.ll == Point(3, 2)
        assert r.ur == Point(105, 4)
        assert r.width == 102
        assert r.height == 2



from __future__ import annotations

from typing import TYPE_CHECKING

from vfbLib.truetype import TT_COMMAND_CONSTANTS, TT_COMMANDS

from FL.helpers.FLList import FLList
from FL.objects.Point import Point
from FL.objects.TTHCommand import TTHCommand

if TYPE_CHECKING:
    from FL.objects.Font import Font
    from FL.objects.Glyph import Glyph
    from FL.objects.Hint import Hint
    from FL.objects.TTHPoint import TTHPoint
    from FL.objects.TTHProblem import TTHProblem


class TTH:
    """
    TTH - class to represent TrueType points
    """

    __slots__ = [
        "_font",
        "_glyph",
        "_top_zones",
        "_bottom_zones",
        "_base_top_zones",
        "_base_bottom_zones",
        "_hstems",
        "_vstems",
        "_base_hstems",
        "_base_vstems",
        "_zero_point",
        "_upm",
        "_ppm",
        "_outline",
        "_base_outline",
        "_commands",
        "_problems",
    ]

    # Constructor

    def __init__(self, g: Glyph | None = None, f: Font | None = None) -> None:
        """
        TTH()                - generic constructor, creates an empty TTH
        TTH(Glyph g)         - creates TTH, assigns 'g' as glyph and 'parent' of 'g' as
                               font
        TTH(Glyph g, Font f) - creates TTH, assigns 'g' as glyph and 'f' as font

        Args:
            g (Glyph | None, optional): _description_. Defaults to None.
            f (Font | None, optional): _description_. Defaults to None.
        """
        self.glyph = None
        self.font = None

        if g is not None:
            self.glyph = g
            if f is None:
                self.font = g.parent
            else:
                self.font = f

        self._top_zones: list[Hint] = FLList()
        self._bottom_zones: list[Hint] = FLList()
        self._base_top_zones: list[Hint] = FLList()
        self._base_bottom_zones: list[Hint] = FLList()
        self.hstems: list[int] = FLList()
        self.vstems: list[int] = FLList()
        self.base_hstems: list[int] = FLList()
        self.base_vstems: list[int] = FLList()
        self.zero_point = Point(0, 0)
        # self.zero_point._parent = self  # Needed?
        self.upm = 1000
        self.ppm = 0  # FL returns a random number
        self._outline: list[TTHPoint] = FLList()
        self._base_outline: list[TTHPoint] = FLList()
        self._commands: list[TTHCommand] = FLList()
        self._problems: list[TTHProblem] = FLList()

    def _raise(self, attr: str) -> None:
        raise RuntimeError(
            f'attempt to write to read-only attribute "{attr}" of class TTH'
        )

    def __repr__(self) -> str:
        return f"<TTH: Glyph: {self.glyph}, Font: {self.font}>"

    # Attributes

    @property
    def font(self) -> Font | None:
        return self._font

    @font.setter
    def font(self, value: Font | None) -> None:
        self._font = value

    @property
    def glyph(self) -> Glyph | None:
        return self._glyph

    @glyph.setter
    def glyph(self, value: Glyph | None) -> None:
        self._glyph = value

    @property
    def top_zones(self) -> list[Hint]:
        return self._top_zones

    @top_zones.setter
    def top_zones(self, value: list[Hint]) -> None:
        self._raise("top_zones")

    @property
    def bottom_zones(self) -> list[Hint]:
        return self._bottom_zones

    @bottom_zones.setter
    def bottom_zones(self, value: list[Hint]) -> None:
        self._raise("bottom_zones")

    @property
    def base_top_zones(self) -> list[Hint]:
        return self._base_top_zones

    @base_top_zones.setter
    def base_top_zones(self, value: list[Hint]) -> None:
        self._raise("base_top_zones")

    @property
    def base_bottom_zones(self) -> list[Hint]:
        return self._base_bottom_zones

    @base_bottom_zones.setter
    def base_bottom_zones(self, value: list[Hint]) -> None:
        self._raise("base_bottom_zones")

    @property
    def hstems(self) -> list[int]:
        return self._hstems

    @hstems.setter
    def hstems(self, value: list[int]) -> None:
        # TODO: Must set FLList()?
        self._hstems = value

    @property
    def vstems(self) -> list[int]:
        return self._vstems

    @vstems.setter
    def vstems(self, value: list[int]) -> None:
        # TODO: Must set FLList()?
        self._vstems = value

    @property
    def base_hstems(self) -> list[int]:
        return self._base_hstems

    @base_hstems.setter
    def base_hstems(self, value: list[int]) -> None:
        # TODO: Must set FLList()?
        self._base_hstems = value

    @property
    def base_vstems(self) -> list[int]:
        return self._base_vstems

    @base_vstems.setter
    def base_vstems(self, value: list[int]) -> None:
        # TODO: Must set FLList()?
        self._base_vstems = value

    @property
    def zero_point(self) -> Point:
        return self._zero_point

    @zero_point.setter
    def zero_point(self, value: Point) -> None:
        self._zero_point = value

    @property
    def upm(self) -> int:
        return self._upm

    @upm.setter
    def upm(self, value: int) -> None:
        self._upm = value

    @property
    def ppm(self) -> int:
        return self._ppm

    @ppm.setter
    def ppm(self, value: int) -> None:
        self._ppm = value

    @property
    def outline(self) -> list[TTHPoint]:
        return self._outline

    @outline.setter
    def outline(self, value: list[TTHPoint]) -> None:
        raise RuntimeError("class TTH has no attribute outline")

    @property
    def base_outline(self) -> list[TTHPoint]:
        return self._base_outline

    @base_outline.setter
    def base_outline(self, value: list[TTHPoint]) -> None:
        raise RuntimeError("class TTH has no attribute base_outline")

    @property
    def commands(self) -> list[TTHCommand]:
        return self._commands

    @commands.setter
    def commands(self, value: list[TTHCommand]) -> None:
        self._raise("commands")

    @property
    def problems(self) -> list[TTHProblem]:
        return self._problems

    @problems.setter
    def problems(self, value: list[TTHProblem]) -> None:
        self._raise("problems")

    # Operations

    def __len__(self) -> int:
        """
        returns number of commands
        """
        return len(self.commands)

    def __getitem__(self, index: int) -> TTHCommand:
        """
        accesses TTCommand list
        """
        return self.commands[index]

    # Methods

    def Init(self, g_or_f: Glyph | Font, f: Font | None = None) -> None:
        raise NotImplementedError

    def Initoutline(self) -> None:
        raise NotImplementedError

    def SetPPM(self, ppm: int) -> None:
        self.ppm = ppm

    def ResetProgram(self, direction: int | None = None) -> None:
        raise NotImplementedError

    def LoadProgram(self, g: Glyph | None = None) -> None:
        # Initialize TTH object with Glyph first or specify it explicitly
        if g is None:
            if self.glyph is None:
                raise ValueError(
                    "TTH.LoadProgram without glyph argument will send FontLab into an "
                    "infinite loop (if the TTH was not initialized with a Glyph)."
                )
            g = self.glyph

        # TODO: Does calling this with a glyph replace the current glyph?

        self.commands.clear()
        for i in g._tth:
            cmd = i["cmd"]
            cmd_int = TT_COMMAND_CONSTANTS[cmd]
            params = [i["params"][param] for param in TT_COMMANDS[cmd_int]["params"]]
            tthcmd = TTHCommand(cmd_int, *params)
            self.commands.append(tthcmd)

    def SaveProgram(self, g: Glyph | None = None) -> None:
        if g is None:
            if self.glyph is None:
                raise ValueError(
                    "TTH.LoadProgram without glyph argument will send FontLab into an "
                    "infinite loop (if the TTH was not initialized with a Glyph)."
                )
            g = self.glyph

        # TODO: Does calling this with a glyph replace the current glyph?

        raise NotImplementedError

    def SortProgram(self) -> None:
        raise NotImplementedError

    def RunProgram(self, i: int | None = None, j: int | None = None) -> None:
        raise NotImplementedError

    def RunCommand(self, tthcommand: TTHCommand) -> None:
        raise NotImplementedError

    def TestProgram(self) -> int:
        # Returns some kind of error code? 1 for an empty program e.g.
        raise NotImplementedError

    def BuildFromLinks(self, g: Glyph | None = None) -> None:
        if g is None:
            raise ValueError(
                "TTH.LoadProgram without glyph argument will send FontLab into an "
                "infinite loop (if the TTH was not initialized with a Glyph)."
            )
        raise NotImplementedError

    def ResetFinalDeltas(self, direction: int | None = None) -> None:
        raise NotImplementedError

    def ResetAllDeltas(self, direction: int | None = None) -> None:
        raise NotImplementedError

    def LoadFromTextFile(self, filename: str) -> int:
        """
        Load TTH from a text file.

        Args:
            filename (str): The path and filename of the text file.

        Returns:
            int: 0 on success, -1 on error.
        """
        raise NotImplementedError

    def SaveToTextFile(self, filename: str) -> None:
        """
        Save the TTH to a text file at path `filename`.

        Args:
            filename (str): The path and filename of the text file.

        Example file:
        ----
        34 POINTS
        439 412 17
        421 419 2
        382 427 2
        363 427 1
        318 427 2
        255 384 2
        220 322 2
        205 254 2
        205 225 1
        205 181 2
        248 135 2
        296 135 1
        320 135 2
        364 144 2
        388 152 1
        385 11 1
        354 2 2
        282 -7 2
        250 -7 1
        185 -7 2
        97 30 2
        51 90 2
        33 158 2
        33 194 1
        33 259 2
        73 390 2
        154 500 2
        287 569 2
        374 569 1
        405 569 2
        459 562 2
        482 554 1
        0 -17 17
        441 -17 17

        5 BOTTOMZONES
        344 10
        0 15
        -83 15
        -171 15
        -233 15

        10 TOPZONES
        815 8
        792 10
        730 15
        705 15
        675 15
        650 15
        600 15
        520 15
        497 15
        470 15

        19 HSTEMS
        140
        152
        158
        171
        220
        116
        85
        97
        109
        123
        130
        164
        145
        180
        200
        22222
        44444
        100
        44

        25 VSTEMS
        182
        195
        220
        152
        208
        91
        128
        139
        66
        100
        110
        117
        127
        135
        144
        148
        169
        179
        189
        200
        215
        225
        22222
        55555
        54

        0 PROGRAM
        ----

        """
        raise NotImplementedError



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from FL.fake.KerningClass import KerningClass

if TYPE_CHECKING:
    from pathlib import Path

    from FL.objects.Font import Font


logger = logging.getLogger(__name__)


class FakeKerning:
    def __init__(self, font: Font | None = None) -> None:
        self._font = font
        # self.num_masters: int = 0
        # self.upm: int | None = None
        self.reset_classes()
        self.reset_pairs()

    def __len__(self) -> int:
        """
        Return the number of key pairs.

        Returns:
            int: The number of kerning pairs. Key and exception pairs only.
        """
        return len(self.kerning)

    def _expand(self) -> None:
        """
        Expand class-based kerning to flat kerning.
        """
        if self._font is None:
            raise ValueError

        self.flat_kerning = []
        cc = []  # Class-class pairs
        gc = []  # Glyph-class pairs
        cg = []  # Class-glyph pairs
        gg = []  # Glyph-glyph pairs
        for g in self._font.glyphs:
            for kerning_pair in g.kerning:
                value = kerning_pair.value
                if value == 0:
                    continue
                L = g.name
                R = self._font.glyphs[kerning_pair.key].name
                l_class = self.classes_left.get(L, None)
                r_class = self.classes_right.get(R, None)
                if l_class:
                    if r_class:
                        cc.append((l_class, r_class, value))
                    else:
                        cg.append((l_class, R, value))
                else:
                    if r_class:
                        gc.append((L, r_class, value))
                    else:
                        gg.append((L, R, value))

        pairs = {}
        for l_class, r_class, value in cc:
            for L in [l_class.keyglyph] + list(l_class.glyphs):
                for R in [r_class.keyglyph] + list(r_class.glyphs):
                    pairs[(L, R)] = value
        for l_class, R, value in cg:
            for L in [l_class.keyglyph] + list(l_class.glyphs):
                pairs[(L, R)] = value
        for L, r_class, value in gc:
            for R in [r_class.keyglyph] + list(r_class.glyphs):
                pairs[(L, R)] = value
        for L, R, value in gg:
            pairs[(L, R)] = value
        self.flat_kerning = sorted([(g[0], g[1], v) for g, v in pairs.items()])
        logger.info("Expanded kerning: %i pairs." % len(self.flat_kerning))

    def expand(self) -> None:
        if self._font is None:
            logger.error("You need to supply a font before you can expand the kerning")
            return

        self.import_classes_from_font(self._font)
        self._expand()

    def export_afm(self, file_path: Path, expand: bool = True) -> None:
        """
        Export kerning data to an AFM file at `file_path`.

        Args:
            file_path (Path): The path to the AFM file.
            expand (bool, optional): Whether to expand class kerning before export.
                If `False`, only key and exception pairs will be exported. Defaults to
                True.
        """
        if expand:
            self.expand()
        raise NotImplementedError

    def export_flc(self, file_path: Path) -> None:
        raise NotImplementedError

    def import_afm(self, file_path: Path, master_index: int = 0) -> None:
        """
        Import kerning data from an AFM file at `file_path`.

        Args:
            file_path (Path): The path to the AFM file.
            master_index (int, optional): _description_. Defaults to 0.
        """
        raise NotImplementedError

    def import_classes_from_font(self, font: Font) -> None:
        """
        Import the kerning classes from a font.

        Args:
            font (Font): The font to import from.
        """
        self.reset_classes()
        for i, flclass in enumerate(font.classes):
            if not flclass.strip().startswith("_"):
                # Not a kerning class
                continue

            kc = KerningClass(fromFLClass=flclass)

            if len(kc) == 0:
                print(f"Skipping empty class: {kc}")
                continue

            if kc.keyglyph is None:
                print(f"Skipping class without keyglyph: {kc.name}")
                continue

            # Override the sides deduced from the class name
            sides = ""
            if font.GetClassLeft(i):
                sides += "L"
            if font.GetClassRight(i):
                sides += "R"
            kc.sides = sides

            if "L" in kc.sides:
                self.classes_left[kc.keyglyph] = kc

            if "R" in kc.sides:
                self.classes_right[kc.keyglyph] = kc

    def import_flc(self, file_path: Path) -> None:
        """
        Import the kerning classes from a FLC file.

        Args:
            file_path (Path): The path to the FLC file.
        """
        self.reset_classes()
        with open(file_path, "r") as flc:
            class_name = None
            class_glyphs = None
            class_sides = None
            for i, line in enumerate(flc):
                line = line.strip()
                if line.startswith("%%CLASS "):
                    class_name = line.split()[-1]
                if line.startswith("%%GLYPHS "):
                    _, class_glyphs = line.split(maxsplit=1)
                if line.startswith("%%KERNING"):
                    class_sides = line.split()[1]
                if line.startswith("%%END"):
                    if class_name is None:
                        logger.error(f"Skipping class without name in line {i}")
                        class_name = None
                        class_glyphs = None
                        class_sides = None
                        continue

                    if class_sides is None:
                        # Classes without side don't kern in FL
                        # Or they may be non-kerning classes; skip them
                        if class_name.startswith("_"):
                            logger.warning(
                                f"Skipping class without side flags: {class_name}"
                            )
                        class_glyphs = None
                        class_name = None
                        class_sides = None
                        continue

                    kc = KerningClass.fromFontLabClass(f"{class_name}: {class_glyphs}")
                    kc.sides = class_sides
                    if kc.keyglyph is None:
                        # Take the first glyph as keyglyph, if any
                        if kc.glyphs:
                            kc.keyglyph = kc.glyphs[0]
                            logger.warning(
                                f"Class without keyglyph: {class_name}, assuming first "
                                f"glyph is key ({kc.keyglyph})"
                            )
                        else:
                            logger.warning(f"Skipping empty class: {class_name}")
                            class_glyphs = None
                            class_name = None
                            class_sides = None
                            continue

                    if "L" in class_sides:
                        self.classes_left[kc.keyglyph] = kc

                    if "R" in class_sides:
                        self.classes_right[kc.keyglyph] = kc

                    # Clean up for the next class
                    class_glyphs = None
                    class_name = None
                    class_sides = None

    def reset_classes(self) -> None:
        """
        Reset kerning classes to empty dicts.
        """
        self.classes: dict[str, dict[str, KerningClass]] = {"L": {}, "R": {}}
        self.classes_left: dict[str, KerningClass] = self.classes["L"]
        self.classes_right: dict[str, KerningClass] = self.classes["R"]

    def reset_pairs(self) -> None:
        self.kerning: dict[tuple[str, str], list[int]] = {}
        self.flat_kerning: list[tuple[str, str, int]] = []



from __future__ import annotations

from vfbLib.typing import TTStemDict

from FL.fake.Base import Copyable


class TTStem(Copyable):
    """
    TTStem - class to represent TrueType-Stem

    TTStem.__doc__ reports only 'TTStem' ... so use this information with care
    """

    __slots__ = [
        "_name",
        "_width",
        "_ppm1",
        "_ppm2",
        "_ppm3",
        "_ppm4",
        "_ppm5",
        "_ppm6",
    ]

    # Constructor

    def __init__(
        self, ttstem_or_width: TTStem | int | None = None, upm: int | None = None
    ) -> None:
        self._name = ""
        self._width = 0
        self._ppm1 = 0  # Not in API
        self._ppm2 = 0
        self._ppm3 = 0
        self._ppm4 = 0
        self._ppm5 = 0
        self._ppm6 = 0
        if ttstem_or_width is not None:
            if isinstance(ttstem_or_width, TTStem):
                self._copy_constructor(ttstem_or_width)
            else:
                # fill in ppm steps based on width and upm
                if upm is None:
                    raise RuntimeError("Stem is expected in arg 1: TTStem(TTStem)")
                self.fake_recalc_ppms(ttstem_or_width, upm)

    def __repr__(self) -> str:
        return f"<TTStem: ppms: {self.ppm2}, {self.ppm3}, {self.ppm4}, {self.ppm5}, {self.ppm6}>"

    # FakeLab additions

    def fake_deserialize(self, data: TTStemDict) -> None:
        if "stem" not in data:
            # the actual stems entry, with value, name, ppm6
            self.width = data["value"]
            self.name = data["name"]
        ppms: dict[str, int] = data["round"]
        for ppm in range(2, 7):
            if str(ppm) in ppms:
                setattr(self, f"ppm{ppm}", ppms[str(ppm)])
        if "1" in ppms:
            # "TrueType Stem PPEMs 1" has no public API
            self._ppm1 = ppms["1"]

    def fake_recalc_ppms(self, width: int, upm: int) -> None:
        """
        Calculate the ppms at which the scaled stem width will round up to the next pixel

        Args:
            width (int): The width of the stem
            upm (int): The units per em of the font
        """
        self.width = width
        self._ppm1 = 0
        self.ppm2 = int(upm // self.width * 1.5)
        self.ppm3 = int(upm // self.width * 2.5)
        self.ppm4 = int(upm // self.width * 3.5)
        self.ppm5 = int(upm // self.width * 4.5)
        self.ppm6 = int(upm // self.width * 5.5)

    # Attributes

    @property
    def name(self) -> str:
        """
        Name of the stem

        Returns:
            str: _description_
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def width(self) -> int:
        """
        Width of the stem

        Returns:
            int: _description_
        """
        return self._width

    @width.setter
    def width(self, value: int) -> None:
        self._width = value

    # What about ppm1? It is stored in the VFB, but in a different entry, so it is
    # probably a later addition that didn't get added to the Python API.

    @property
    def ppm2(self) -> int:
        """
        From this ppm-size the stem becomes 2 pixels wide

        Returns:
            int: _description_
        """
        return self._ppm2

    @ppm2.setter
    def ppm2(self, value: int) -> None:
        self._ppm2 = value

    @property
    def ppm3(self) -> int:
        """
        From this ppm-size the stem becomes 3 pixels wide

        Returns:
            int: _description_
        """
        return self._ppm3

    @ppm3.setter
    def ppm3(self, value: int) -> None:
        self._ppm3 = value

    @property
    def ppm4(self) -> int:
        """
        From this ppm-size the stem becomes 4 pixels wide

        Returns:
            int: _description_
        """
        return self._ppm4

    @ppm4.setter
    def ppm4(self, value: int) -> None:
        self._ppm4 = value

    @property
    def ppm5(self) -> int:
        """
        From this ppm-size the stem becomes 5 pixels wide

        Returns:
            int: _description_
        """
        return self._ppm5

    @ppm5.setter
    def ppm5(self, value: int) -> None:
        self._ppm5 = value

    @property
    def ppm6(self) -> int:
        """
        From this ppm-size the stem becomes 6 pixels wide

        Returns:
            int: _description_
        """
        return self._ppm6

    @ppm6.setter
    def ppm6(self, value: int) -> None:
        self._ppm6 = value

    # Operations

    # (don't know)

    # Methods

    # (don't know)



import unittest

from FL import Font
from FL.helpers.classList import ClassList


class ClassListTests(unittest.TestCase):
    def test_instantiation_empty(self) -> None:
        c = ClassList()
        assert isinstance(c, ClassList)
        assert c._flags == []

    def test_instantiation(self) -> None:
        c = ClassList(["_A: A'"])
        assert c == ["_A: A'"]
        assert c._flags == [0]

    def test_instantiation_flags(self) -> None:
        c = ClassList(["_A: A'"])
        c.SetClassFlags(0, True, True)
        assert c == ["_A: A'"]
        assert c._flags == [3072]

        # When we pass the old list to __init__, the flags should persist
        c = ClassList(["_A: A'", "_B: B'"], old_list=c)
        assert c._flags == [3072, 0]
        c.SetClassFlags(1, True, False)
        assert c._flags == [3072, 1024]

        # When passing a longer or shorter list, the length of the flags list should be
        # adapted and the existing classes should keep their flags even though the index
        # changes
        c = ClassList(["_B: B'"], old_list=c)
        assert c._flags == [1024]

    def test_add(self) -> None:
        c1 = ClassList(["_A: A'"])
        c2 = ["_O: O'"]
        assert c1 == ["_A: A'"]
        assert c1._flags == [0]
        c3 = c1 + c2
        assert c3 == ["_A: A'", "_O: O'"]
        assert c3._flags == [0, 0]
        assert id(c1) != id(c3)

    def test_iadd(self) -> None:
        c1 = ClassList(["_A: A'"])
        id1 = id(c1)
        c2 = ["_O: O'"]
        assert c1 == ["_A: A'"]
        assert c1._flags == [0]
        c1 += c2
        assert c1 == ["_A: A'", "_O: O'"]
        assert c1._flags == [0, 0]
        assert id1 == id(c1)

    def test_font(self) -> None:
        f = Font()
        f.classes = ["a: a"]
        assert isinstance(f.classes, list)
        assert isinstance(f._classes, ClassList)
        assert f._classes._flags == [0]

    def test_font_add(self) -> None:
        f = Font()
        f.classes += ["b: b"]
        assert f.classes == ["b: b"]
        assert f.GetClassLeft(0) == 0
        assert f.GetClassRight(0) == 0
        f.classes += ["c: c"]
        assert f.classes == ["b: b", "c: c"]
        assert f.GetClassLeft(1) == 0
        assert f.GetClassRight(1) == 0
        assert f.GetClassLeft(2) is None
        assert f.GetClassRight(2) is None

    def test_font_iadd(self) -> None:
        f = Font()
        f.classes = f.classes + ["b: b"]
        assert f.classes == ["b: b"]
        assert f.GetClassLeft(0) == 0
        assert f.GetClassRight(0) == 0
        f.classes = f.classes + ["c: c"]
        assert f.classes == ["b: b", "c: c"]
        assert f.GetClassLeft(1) == 0
        assert f.GetClassRight(1) == 0
        assert f.GetClassLeft(2) is None
        assert f.GetClassRight(2) is None

    def test_font_append(self) -> None:
        # Append has no effect
        f = Font()
        f.classes.append("b: b")
        assert len(f.classes) == 0

    def test_font_extend(self) -> None:
        # Extend has no effect
        f = Font()
        f.classes.extend(["b: b"])
        assert len(f.classes) == 0

    def test_font_insert(self) -> None:
        # Insert has no effect
        f = Font()
        f.classes.insert(0, "b: b")
        assert len(f.classes) == 0



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from FL.fake.Base import Copyable
from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph
    from FL.objects.Matrix import Matrix

# Node type
nLINE = 1
nMOVE = 17
nCURVE = 35
nOFF = 65

# Alignment
nSHARP = 0
nSMOOTH = 4096  # tangent
nCLOSEPATH = 8192  # ? Undocumented: Closepath follows after node
nFIXED = 12288  # curve to curve smooth

vfb2json_node_types = {"line": 1, "move": 17, "curve": 35, "qcurve": 65}
json2vfb_node_types = {nLINE: "line", nMOVE: "move", nCURVE: "curve", nOFF: "qcurve"}

vfb2json_node_conns = {0: nSHARP, 1: nSMOOTH, 2: nCLOSEPATH, 3: nFIXED}
json2vfb_node_conns = {nSHARP: 0, nSMOOTH: 1, nCLOSEPATH: 2, nFIXED: 3}


class Node(Copyable):
    """
    Node() - generic constructor, creates an empty node
    Node(Node) - copy constructor
    Node(integer type, Point p) - creates a Node and assigns type and
    coordinates of the final point
    """

    __slots__ = [
        "alignment",
        "selected",
        "type",
        "_masters_count",
        "_parent",
        "_points",
    ]

    # Constructor

    def __init__(
        self, node_or_type: Node | int | None = None, p: Point | None = None
    ) -> None:
        """
        Node()                      - generic constructor, creates an empty node
        Node(Node)                  - copy constructor
        Node(integer type, Point p) - creates a Node and assigns type and coordinates of
                                      the final point

        Args:
            node_or_type (Node | int | None, optional): _description_. Defaults to None.
            p (Point | None, optional): _description_. Defaults to None.
        """
        # Remove float when setting coords
        self._set_defaults()

        # Process params

        if isinstance(node_or_type, Node):
            assert p is None
            self._copy_constructor(node_or_type)
        elif isinstance(node_or_type, int):
            assert isinstance(p, Point)

            self.type = node_or_type
            p = Point(int(p.x), int(p.y))
            p._parent = self
            self._points = [[p] for _ in range(self._masters_count)]
        # else: Empty node

    def __repr__(self) -> str:
        return f"<Node: type=0x{self.type:x}, x={self.x:g}, y={self.y:g}>"

    # Additions for FakeLab

    def fake_deserialize(self, num_masters: int, data: dict[str, Any]) -> None:
        self._masters_count = num_masters
        self.type = vfb2json_node_types[data["type"]]
        self.alignment = vfb2json_node_conns[data["flags"]]
        self._points = [[] for _ in range(self._masters_count)]
        points = data.get("points", [])
        for master_index in range(num_masters):
            master_points = points[master_index]
            if self.type in (nMOVE, nLINE, nOFF):
                assert len(master_points) == 1
            elif self.type == nCURVE:
                assert len(master_points) == 3
            else:
                raise ValueError(f"Unknown Node type: {self.type}")
            for x, y in master_points:
                self._points[master_index].append(Point(x, y))
        if self.type in (nMOVE, nLINE, nOFF):
            assert len(self.points) == 1
        elif self.type == nCURVE:
            assert len(self.points) == 3
        else:
            raise ValueError(f"Unknown Node type: {self.type}")

    def fake_serialize(self, num_masters: int) -> dict[str, Any]:
        d = {
            "type": json2vfb_node_types[self.type],
            "flags": json2vfb_node_conns[self.alignment],
            "points": [],
        }
        points: list[list[tuple[int, int]]] = [[] for _ in range(num_masters)]
        for master_index, master_points in enumerate(self._points):
            for p in master_points:
                points[master_index].append((int(p.x), int(p.y)))
        d["points"] = points
        return d

    def fake_update(self, glyph: Glyph | None = None) -> None:
        """
        Is called from FontLab.UpdateFont()
        """
        self._parent: Glyph | None = glyph
        for p in self.points:
            p.fake_update(self)

    # Attributes

    @property
    def parent(self) -> Glyph | None:
        """
        Nodes's parent object, Glyph
        """
        return self._parent

    @property
    def count(self) -> int:
        return len(self)

    @property
    def point(self) -> Point:
        """
        position of the final point of the first master
        """
        return self._points[0][-1]

    @property
    def points(self) -> list[Point]:
        """
        positions of all points of the first master
        """
        return self._points[0]

    @property
    def x(self) -> int:
        return int(self.point.x)

    @property
    def y(self) -> int:
        return int(self.point.y)

    # Operations

    def __len__(self) -> int:
        """
        Return the number of points.
        """
        return len(self._points[0])

    def __getitem__(self, index: int) -> Point:
        """
        Accesses points array of the first master
        """
        return self._points[0][index]

    def __mul__(self, matrix: Matrix) -> Node:
        raise NotImplementedError

    # Methods

    def Assign(
        self, node_or_type: Node | int | None = None, p: Point | None = None
    ) -> None:
        """
        Assigns new values to a Node, refer to constructor for a description of
        possible options
        """
        raise NotImplementedError

    def SetAllLayers(self, pointindex: int, p: Point) -> None:
        """
        Assigns position of the point p to all masters of the point number
        'pointindex'
        """
        raise NotImplementedError

    def Layer(self, masterindex: int) -> list[Point]:
        """
        Returns list of points for the master 'masterindex'
        """
        return self._points[masterindex]

    def Section(self, pointindex: int) -> list[Point]:
        """
        Returns list of points for all layers and point number 'pointindex'
        """
        raise NotImplementedError

    def Shift(self, p: Point, masterindex: int = 0) -> None:
        """
        shifts position of all points for the master 'masterindex'
        """
        raise NotImplementedError

    def ExpandLayer(self, masterindex: int) -> None:
        """
        copies positions of all points in the master 'masterindex' to other masters
        """
        raise NotImplementedError

    def Transform(self, m: Matrix) -> None:
        """
        Applies Matrix transformation to the Node
        """
        raise NotImplementedError

    # Defaults

    def _set_defaults(self) -> None:
        self._parent = None
        self._masters_count = 1

        # type of the node, values are: nMOVE, nLINE, nCURVE or nOFF
        self.type = nLINE

        # type of primitive connection, possible values are:
        # nSHARP, nSMOOTH, nFIXED
        self.alignment = nSHARP

        # True if node is selected
        self.selected = 0
        self._points = [[Point()] for _ in range(self._masters_count)]



import unittest

from FL.objects.Node import Node, nLINE
from FL.objects.Point import Point


class NodeTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        n = Node()
        assert n.parent is None
        assert n.type == nLINE
        assert n.x == 0
        assert n.y == 0

    def test_instantiation_float_point(self) -> None:
        p = Point(1.8, 1.2)
        n = Node(nLINE, p)
        assert n.type == 1
        assert n.x == 1
        assert n.y == 1
        assert n.point.x == 1.0



import unittest

from FL.objects.KerningPair import KerningPair


class KerningPairTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        k = KerningPair()
        assert k.parent is None
        assert k.key == 0
        assert k.value == 0
        assert k.values == [0] * 16



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Font import Font


class NameRecord(Copyable):
    """
    NameRecord - class to represent OpenType name table record

    Unicode-encoding syntax ('/XX' - for Mac name records
                   and '/XXXX' - for Windows and Unicode name records)
    must be used to assign characters with codes above 127.
    Refer to OpenType specification for information about constants used
    for platform, encoding, language and name IDs
    for newline, Use /000D/000A in Win records and /0D in Mac records. (Description
    revised)
    """

    __slots__ = ["_parent", "_nid", "_pid", "_eid", "_lid", "_name"]

    # Constructor

    def __init__(
        self,
        name_record_or_s_or_tup_or_nid: (
            NameRecord | str | tuple[int, int, int, int, str] | int | None
        ) = None,
        pid: int | None = None,
        eid: int | None = None,
        lid: int | None = None,
        s: str | None = None,
    ) -> None:
        self._parent = None
        arg1 = name_record_or_s_or_tup_or_nid
        if isinstance(arg1, NameRecord):
            self._copy_constructor(arg1)
        elif isinstance(arg1, str):
            self.nid = 0
            self.pid = 3
            self.eid = 1
            self.lid = 1033
            self.name = arg1
        elif isinstance(arg1, tuple):
            self.nid, self.pid, self.eid, self.lid, self.name = arg1
        elif isinstance(arg1, int):
            self.nid = arg1
            assert pid is not None
            self.pid = pid
            assert eid is not None
            self.eid = eid
            assert lid is not None
            self.lid = lid
            assert s is not None
            self.name = s

    # Additions for FakeLab

    def fake_serialize(self) -> tuple[int, int, int, int, str]:
        return (self.nid, self.pid, self.eid, self.lid, self.name)

    # Attributes

    @property
    def parent(self) -> Font | None:
        return self._parent

    @property
    def nid(self) -> int:
        """
        name identifier

        Returns:
            int: _description_
        """
        return self._nid

    @nid.setter
    def nid(self, value: int) -> None:
        self._nid = value

    @property
    def pid(self) -> int:
        """
        platform identifier

        Returns:
            int: _description_
        """
        return self._pid

    @pid.setter
    def pid(self, value: int) -> None:
        self._pid = value

    @property
    def eid(self) -> int:
        """
        encoding identifier

        Returns:
            int: _description_
        """
        return self._eid

    @eid.setter
    def eid(self, value: int) -> None:
        self._eid = value

    @property
    def lid(self) -> int:
        """
        language identifier

        Returns:
            int: _description_
        """
        return self._lid

    @lid.setter
    def lid(self, value: int) -> None:
        self._lid = value

    @property
    def name(self) -> str:
        """
        name value

        Returns:
            str: _description_
        """
        # FIXME: What encoding are we dealing with here?
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    # Operations: none

    # Methods: none



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph


class Link(Copyable):
    """
    Link - class to represent link
    """

    __slots__ = ["_node1", "_node2", "_parent"]

    # Constructor

    def __init__(
        self, link_or_index1: Link | int | None = None, index2: int | None = None
    ) -> None:
        self.set_defaults()

        # Process params

        if isinstance(link_or_index1, Link):
            self._copy_constructor(link_or_index1)

        elif isinstance(link_or_index1, int):
            self._node1 = link_or_index1

            if index2 is not None:
                self._node2 = index2

    def __repr__(self) -> str:
        p = self._parent or "orphan"
        return f"<Link: node1={self.node1}, node2={self.node2}, {p}>"

    # Attributes

    @property
    def parent(self) -> Glyph | None:
        """
        Link's parent object, Glyph
        """
        return self._parent

    @property
    def node1(self) -> int:
        """
        indexes of the nodes that are linked: node1
        """
        return self._node1

    @node1.setter
    def node1(self, value: int) -> None:
        self._node1 = value

    @property
    def node2(self) -> int:
        """
        indexes of the nodes that are linked: node2
        """
        return self._node2

    @node2.setter
    def node2(self, value: int) -> None:
        self._node2 = value

    # Methods

    def ToHint(self) -> None:
        """
        Transforms link to Hint (and returns it as a result **wrong**) using parent as
        a source of node coordinates. Parent must exist
        """
        # This does *not* return the hint, but seems to append it to the
        # glyph's hhints or vhints property
        return None

    # Defaults

    def set_defaults(self) -> None:
        self._parent = None
        self._node1 = -1
        self._node2 = -1



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable
from FL.objects.Hint import Hint
from FL.objects.Node import Node

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph


TYPE_HORIZONTAL_HINT = 1
TYPE_VERTICAL_HINT = 2
TYPE_NODE = 255


class Replace(Copyable):
    """
    Replace - class to represent item of hint replacment program

    Replace.__doc__ reports only 'Replace' ... so use this information with care

    To use attribute object, Replase must have valid parent
    To assign a `Node`, Node must not be an orphan
    To assign a `Hint`, Hint must not be an orphan
    """

    """
    Example:
    for r in fl.glyph.replace_table:
        print r
    <Replace: HorizontalHint number 2, parent: "a">
    <Replace: VerticalHint number 0, parent: "a">
    <Replace: HorizontalHint number 0, parent: "a">
    <Replace: HintReplace at node 6, parent: "a">
    <Replace: HorizontalHint number 1, parent: "a">
    <Replace: VerticalHint number 0, parent: "a">
    <Replace: HorizontalHint number 2, parent: "a">
    """

    # Constructor

    def __init__(
        self,
        replace_or_hint_or_node_or_type: Replace | Hint | Node | int | None = None,
        index: int | None = None,
    ) -> None:
        """
        Replace()             - generic constructor, creates an empty replace item
        Replace(Replace)      - copy constructor
        Replace(Hint)
        Replace(Node)
        Replace(Int type,Int index)

        Args:
            replace_or_hint_or_node_or_type (Replace | Hint | Node | int | None, optional): _description_. Defaults to None.
            index (int | None, optional): _description_. Defaults to None.
        """
        self._parent: Glyph | None = None
        self._type = 0
        self._index = 0
        arg1 = replace_or_hint_or_node_or_type
        if arg1 is not None:
            if isinstance(arg1, Replace):
                self._copy_constructor(arg1)
            elif isinstance(arg1, Hint):
                # FIXME: How to determine if a hint is horizontal or vertical?
                self._type = TYPE_HORIZONTAL_HINT  # or TYPE_VERTICAL_HINT
            elif isinstance(arg1, Node):
                self._type = TYPE_NODE
            elif isinstance(arg1, int):
                self.type = arg1
                if isinstance(index, int):
                    self.index = index
            else:
                raise TypeError

    def __repr__(self) -> str:
        tp = {
            1: "HorizontalHint",
            2: "VerticalHint",
            # 255: "HintReplace",
        }

        parent = "orphan" if self.parent is None else f'parent: "{self.parent.name}"'

        if self.type == TYPE_NODE:
            return f"<Replace: HintReplace at node {self.index}, {parent}>"
        elif self.type in (TYPE_HORIZONTAL_HINT, TYPE_VERTICAL_HINT):
            return f"<Replace: {tp.get(self.type)} number {self.index}, {parent}>"
        else:
            return f"<Replace: Unknown replace type, index= {self.index}, {parent}>"

    @property
    def parent(self) -> Glyph | None:
        """
        Replace's parent object, `Glyph`

        Returns:
            Glyph | None: _description_
        """
        return self._parent

    @property
    def type(self) -> int:
        return self._type

    @type.setter
    def type(self, value: int) -> None:
        self._type = value

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, value: int) -> None:
        self._index = value

    # Operations: unknown

    # Methods: unknown



from __future__ import annotations

import unittest

from FL.helpers.FLList import adjust_list


class FLListTests(unittest.TestCase):
    def test_adjust_list_equal(self):
        seq = [0, 0, 0, 0]
        expected = [0, 0, 0, 0]
        adjust_list(seq, 4)
        assert seq == expected

    def test_adjust_list_empty(self):
        seq = []
        expected = [None, None, None]
        adjust_list(seq, 3)
        assert seq == expected

    def test_adjust_list_empty_value(self):
        seq = []
        expected = [0, 0, 0]
        adjust_list(seq, 3, 0)
        assert seq == expected

    def test_adjust_list_no_value(self):
        seq = [0]
        expected = [0, 0, 0]
        adjust_list(seq, 3)
        assert seq == expected

    def test_adjust_list_value(self):
        seq = [1]
        expected = [1, 0, 0]
        adjust_list(seq, 3, 0)
        assert seq == expected

    def test_adjust_list_shorter_no_value(self):
        seq = [0, 0, 0]
        expected = [0]
        adjust_list(seq, 1)
        assert seq == expected

    def test_adjust_list_shorter_value(self):
        seq = [1, 0, 0]
        expected = [1]
        adjust_list(seq, 1, 0)  # value is ignored
        assert seq == expected



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph


class AuditRecord:
    """
    AuditRecord - class to represent Font Audit error record
    """

    __slots__ = ["_parent", "_x", "_y", "_id", "_description", "_index"]

    # Constructor

    def __init__(self, parent: Glyph, _called_from_FL: bool = False) -> None:
        """
        There is no explicit constructor for object of this type.

        Note that objects of this class cannot be created explicitly, they are only
        generated as a result of a Glyph.Audit() operation

        Args:
            parent (Glyph): The parent of the AuditRecord, a glyph.
        """
        self._parent: Glyph = parent
        self._x = 0
        self._y = 0
        self._id = ""
        self._description = ""
        self._index = -1
        if not _called_from_FL:
            raise RuntimeError(
                "There is no explicit constructor for object of this type"
            )

    # Additions for FakeLab

    @staticmethod
    def fake_create(parent: Glyph) -> AuditRecord:
        a = AuditRecord.__new__(AuditRecord)
        a._parent = parent
        a._x = 0
        a._y = 0
        a._id = ""
        a._description = ""
        a._index = -1
        return a

    # Attributes

    @property
    def parent(self) -> Glyph:
        """
        The audit record's parent object.

        Returns:
            Glyph: The parent glyph.
        """
        return self._parent

    @property
    def position(self) -> Point:
        """
        The position of the audit mark as a point.

        Returns:
            Point: The point.
        """
        return Point(self._x, self._y)

    @property
    def p(self) -> Point:
        """
        The position of the audit mark as a point.

        Returns:
            Point: The point.
        """
        return Point(self._x, self._y)

    @property
    def id(self) -> str:
        """
        The error name.

        Returns:
            str: The name.
        """
        return self._id

    @property
    def description(self) -> str:
        """
        The error description.

        Returns:
            str: The description.
        """
        return self._description

    @property
    def index(self) -> int:
        """
        The node index in the glyph to which error is attached.

        Returns:
            int: The node index.
        """
        return self._index

    # Methods

    def CanBeFixed(self) -> bool:
        """
        Returns:
            bool: True if this error can be automatically fixed.
        """
        raise NotImplementedError

    def Repair(self) -> None:
        """
        Try to automatically fix error.
        """
        raise NotImplementedError



import unittest

from FL.objects.TTStem import TTStem


class TTStemTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        # No args
        t = TTStem()
        assert t.name == ""
        assert t.width == 0
        assert t.ppm2 == 0
        assert t.ppm3 == 0
        assert t.ppm4 == 0
        assert t.ppm5 == 0
        assert t.ppm6 == 0

    def test_instantiation_copy(self) -> None:
        # Use the copy constructor
        t0 = TTStem()
        t0.name = "Hello"
        t0.width = 120
        t0.ppm2 = 8
        t0.ppm3 = 10
        t0.ppm4 = 14
        t0.ppm5 = 34
        t0.ppm6 = 45

        t = TTStem(t0)
        assert t.name == "Hello"
        assert t.width == 120
        assert t.ppm2 == 8
        assert t.ppm3 == 10
        assert t.ppm4 == 14
        assert t.ppm5 == 34
        assert t.ppm6 == 45

    def test_instantiation_args(self) -> None:
        # Initialize with width and upm
        t = TTStem(100, 1000)
        assert t.name == ""
        assert t.width == 100
        assert t.ppm2 == 15
        assert t.ppm3 == 25
        assert t.ppm4 == 35
        assert t.ppm5 == 45
        assert t.ppm6 == 55



from __future__ import annotations


class TTHProblem:
    """
    TTHProblem - class to represent TrueType problems
    """

    __slots__ = ["_id", "_command", "_direction"]

    # Constructor

    def __init__(self, _called_from_FL: bool = False) -> None:
        """
        Class `TTHProblem` is auxiliary and cannot be created explicitly
        """
        # Those values have not been checked against FL
        self._id = 0
        self._command = 0
        self._direction = 0
        if not _called_from_FL:
            raise RuntimeError(
                "Class TTHProblem is auxiliary and cannot be created explicitly"
            )

    def __repr__(self) -> str:
        return f"<TTHProblem: Id: {self.id}, Command: {self.command}, Direction: {self.direction}>"

    # Additions for FakeLab

    @staticmethod
    def fake_create() -> TTHProblem:
        p = TTHProblem.__new__(TTHProblem)
        return p

    # Attributes

    @property
    def id(self) -> int:
        return self._id

    @property
    def command(self) -> int:
        return self._command

    @property
    def direction(self) -> int:
        return self._direction



from __future__ import annotations

from collections import UserList
from copy import copy
from typing import Any, Iterable, SupportsIndex


class ClassList(UserList[str]):
    """
    A list of OpenType classes as strings. It keeps track of the class flags.
    """

    __slots__ = ["_flags", "_kerning_flags", "_metrics_flags", "_names"]

    def __init__(self, iterable: Iterable[str] | None = None) -> None:
        super().__init__(iterable)
        self._flags = [0 for _ in iterable or []]
        self._kerning_flags: dict[str, list[int]] = {}
        self._metrics_flags: dict[str, list[int]] = {}
        self._names: tuple[str, ...] = tuple()
        self._update_flags()

    # Internal

    def _get_class_name(self, class_string: str) -> str:
        if ":" not in class_string:
            raise ValueError
        name, _contents = class_string.split(":", 1)
        return name.strip()

    def _update_names(self) -> None:
        # Mapping from name to class index?
        self._names = tuple(
            [
                name
                for name in [
                    self._get_class_name(class_string) for class_string in self.data
                ]
            ]
        )

    def _update_flags(self) -> None:
        # Match classes from the old list so they can keep their flags

        # Reset flags
        self._flags = [0] * len(self.data)

        if not self._flags:
            return

        # Make a list of all class types
        self._update_names()

        # Set kerning flags
        for name, flags in self._kerning_flags.items():
            class_index = self._names.index(name)
            self._flags[class_index] = flags[0]

        # Set metrics flags
        for name, flags in self._metrics_flags.items():
            class_index = self._names.index(name)
            self._flags[class_index] = flags[1]

    @property
    def fake_metrics_flags(self) -> dict[str, list[int]]:
        # Called from FL.vfb.writer
        # Serialize
        pass

    @fake_metrics_flags.setter
    def fake_metrics_flags(self, value: dict[str, list[int]]) -> None:
        # Called from FL.vfb.reader
        self._metrics_flags = value

    @property
    def fake_kerning_flags(self) -> dict[str, list[int]]:
        # Called from FL.vfb.writer
        pass

    @fake_kerning_flags.setter
    def fake_kerning_flags(self, value: dict[str, list[int]]) -> None:
        # Called from FL.vfb.reader
        self._kerning_flags = value

    def fake_set_classes(self, classes: list[str]) -> None:
        # Called from FL.vfb.reader
        # Carries over the flags when setting the value
        self.data = classes
        self._update_flags()

    # Operations

    def __add__(self, item: Any) -> ClassList:
        result = ClassList(self.data)
        result._kerning_flags = copy(self._kerning_flags)
        result._metrics_flags = copy(self._metrics_flags)
        result += item
        return result

    def __iadd__(self, item: Any) -> ClassList:
        # self._flags.append(0)
        self.data.__iadd__(item)
        return self

    def __setitem__(self, index: SupportsIndex | int | Any, item: Any) -> None:
        # Not implemented, it does nothing
        pass

    def append(self, item: Any) -> None:
        # WTF ... append does not work on the classes list
        pass

    def extend(self, iterable: Iterable) -> None:
        # extend does not work either
        pass

    def insert(self, index: int, item: Any) -> None:
        # You guessed it
        pass

    # Methods called by the Font

    def GetClassLeft(self, class_index: int) -> int | None:
        if class_index >= len(self.data) or class_index < 0:
            return None

        flags = self._flags[class_index]
        if flags & 1:
            # Metrics class
            return 0

        return int(bool(flags & 1024))

    def GetClassRight(self, class_index: int) -> int | None:
        if class_index >= len(self.data) or class_index < 0:
            return None

        flags = self._flags[class_index]
        if flags & 1:
            # Metrics class
            return 0

        return int(bool(flags & 2048))

    def GetClassMetricsFlags(self, class_index: int) -> tuple | None:
        if class_index >= len(self.data) or class_index < 0:
            return None

        flags = self._flags[class_index]
        if not flags & 1:
            return (0, 0, 0)
        return (
            int(bool(flags & 1024)),  # L
            int(bool(flags & 2048)),  # R
            int(bool(flags & 4096)),  # W
        )

    def SetClassFlags(
        self,
        class_index: int,
        left_lsb: bool | int,
        right_rsb: bool | int,
        width: bool | int | None = None,
    ) -> None:
        if class_index >= len(self.data) or class_index < 0:
            return None

        value = 0

        if left_lsb:
            value += 1024
        if right_rsb:
            value += 2048
        if width:
            value += 4096

        class_name = self._names[class_index]
        if width is None:
            # Kerning class
            self._kerning_flags[class_name] = [value, 0]
        else:
            # Must be a metrics class
            value += 1
            self._metrics_flags[class_name] = [0, value, 0]

        self._flags[class_index] = value



from __future__ import annotations

from FL.fake.Base import Copyable


class TTGasp(Copyable):
    """
    This class is completely undocumented
    """

    __slots__ = ["_ppm", "_behavior"]

    def __init__(self, ttgasp_or_ppm: TTGasp | int | None = None, behavior: int = 0):
        """
        A gasp table record.

        Args:
            ttgasp_or_ppm (TTGasp | int | None, optional): Copy attributes from
                `TTGasp`, set the ppm, or create a default gasp table record. Defaults
                to None.
            behavior (int, optional): The behavior flags for the ppm range. Must be used
                together with the ppm integer in the first argument. Defaults to 0.

        Raises:
            RuntimeError: When the arguments don't match the expectations.
        """
        self.ppm = 1000
        self.behavior = behavior
        if ttgasp_or_ppm is not None:
            if isinstance(ttgasp_or_ppm, TTGasp):
                self._copy_constructor(ttgasp_or_ppm)
            else:
                if not isinstance(behavior, int):
                    raise RuntimeError("Gasp is expected in arg 1: TTGasp(TTGasp)")
                self.ppm = ttgasp_or_ppm
                self.behavior = behavior

    def __repr__(self) -> str:
        return f"<TTGasp: {self.ppm}, {self.behavior}, Orphan>"

        # Or if parent is present:
        # f"<TTGasp: {self.ppm}, {self.behavior}, Reference>"

    # Additions for FakeLab

    def fake_deserialize(self, data: dict[str, int]) -> None:
        self.ppm = data["maxPpem"]
        self.behavior = data["flags"]

    def fake_serialize(self) -> dict[str, int]:
        return {"maxPpem": self.ppm, "flags": self.behavior}

    # Attributes

    @property
    def ppm(self) -> int:
        """
        Upper limit of range, in PPEM

        Returns:
            int: The maximum ppm for the gasp record
        """
        return self._ppm

    @ppm.setter
    def ppm(self, value: int) -> None:
        self._ppm = value

    @property
    def behavior(self) -> int:
        """
        Flags describing desired rasterizer behavior.

        Returns:
            int: The flags.

        0x0001 - Use gridfittig
        0x0002 - Use grayscale rendering

        Those additional flags are defined in version 1 of the gasp table specification,
        but they can not be accessed from the FontLab 5 UI:

        0x0004 - Use gridfitting with ClearType symmetric smoothing
        0x0008 - Use smoothing along multiple axes with ClearType®
        """
        # TODO: What happens if we set the unsupported flags via Python?
        return self._behavior

    @behavior.setter
    def behavior(self, value: int) -> None:
        self._behavior = value



from __future__ import annotations

# FLdict

## Canvas

### Brush-Style
cBRUSH_NULL = 1
cBRUSH_SOLID = 0

### Draw-Style
cDS_BLACK = 1
cDS_NOTOR = 2
cDS_NOTCOPY = 4
cDS_NOT = 6
cDS_XOR = 7
cDS_NOTAND = 8
cDS_AND = 9
cDS_NOTXOR = 10
cDS_COPY = 13
cDS_OR = 15
cDS_WHITE = 16

### Pen-Style
cPEN_DASH = 3
cPEN_DOT = 2
cPEN_NULL = 1
cPEN_SOLID = 0

## Colors
cRGB_BLACK = 0
cRGB_BLUE = 16711680
cRGB_GRAY = 8421504
cRGB_GREEN = 65280
cRGB_LTGRAY = 12632256
cRGB_RED = 255
cRGB_WHITE = 16777215
cRGB_YELLOW = 8454143

## Dialog

# -> see Dialog class

### ControlTypes
EDITCONTROL = 0
STATICCONTROL = 1
CHECKBOXCONTROL = 2
BUTTONCONTROL = 3
LISTCONTROL = 4
CHOICECONTROL = 5
PREVIEWCONTROL = 6

### ControlStyles
STYLE_BUTTON = 268435456
STYLE_CHECKBOX = 268435459
STYLE_CHOICE = 278921219
STYLE_CUSTOM = 276824064
STYLE_EDIT = 276824064
STYLE_FRAME = 268435474
STYLE_LABEL = 268435456
STYLE_LIST = 278921216

### ControlStyleModifier
cTO_BOTTOM = 8
cTO_CENTER = 1
cTO_LEFT = 0
cTO_RIGHT = 2
cTO_TOP = 0
cTO_VCENTER = 4

### ControlAlignment
aALIGN = -100
aAUTO = -105
aIDENT = -102
aIDENT2 = -103
aIDENT3 = -104
aNEXT = -101
aSAME = -106


## Message

OK = 1
Cancel = 2


## Node
# -> see Node class

# ### Type
# nLINE = 1
# nMOVE = 17
# nCURVE = 35
# nOFF = 65

# ### Alignment
# nSHARP = 0
# nSMOOTH = 4096
# nFIXED = 12288


## Transformation

TR_CODE_SHIFT = 0
TR_CODE_MIRROR = 1
TR_CODE_SCALE = 2
TR_CODE_ROTATE = 3
TR_CODE_SLANT = 4
TR_CODE_REMOVE = 5
TR_CODE_AUTOHINT = 6
TR_CODE_TT_AUTOHINT = 7
TR_CODE_AUTOREPLACE = 8
TR_CODE_TOLINKS = 9
TR_CODE_TOHINTS = 10
TR_CODE_DECOMPOSE = 11
TR_CODE_CONVERT3 = 12
TR_CODE_CONVERT2 = 13
TR_CODE_REVERSE_ALL = 14
TR_CODE_ALIGNMENT = 15
TR_CODE_EXTREMES = 16
TR_CODE_OVERLAP = 17
TR_CODE_WIDTH = 18
TR_CODE_BEARING = 19
TR_CODE_ALIGNWIDTH = 20
TR_CODE_AUTOSPACING = 21
TR_CODE_BOLD = 22
TR_CODE_COLLEGE = 23
TR_CODE_SHADOW = 24
TR_CODE_3D = 25
TR_CODE_3DROTATE = 26
TR_CODE_GRADIENT = 27
TR_CODE_RANDOM = 28
TR_CODE_ENVELOPE = 29
TR_CODE_MAKELAYER = 30
TR_CODE_OPTIMIZE = 31
TR_CODE_DROP_TTH = 32
TR_CODE_ADJUST = 33


SS_BITMAP = 14
SS_BLACKFRAME = 7
SS_BLACKRECT = 4
SS_CENTER = 1
SS_CENTERIMAGE = 512
SS_ELLIPSISMASK = 49152
SS_ENDELLIPSIS = 16384
SS_ENHMETAFILE = 15
SS_ETCHEDFRAME = 18
SS_ETCHEDHORZ = 16
SS_ETCHEDVERT = 17
SS_GRAYFRAME = 8
SS_GRAYRECT = 5
SS_ICON = 3
SS_LEFT = 0
SS_LEFTNOWORDWRAP = 12
SS_NOPREFIX = 128
SS_NOTIFY = 256
SS_OWNERDRAW = 13
SS_PATHELLIPSIS = 32768
SS_REALSIZEIMAGE = 2048
SS_RIGHT = 2
SS_RIGHTJUST = 1024
SS_SIMPLE = 11
SS_SUNKEN = 4096
SS_TYPEMASK = 31
SS_USERITEM = 10
SS_WHITEFRAME = 9
SS_WHITERECT = 6
SS_WORDELLIPSIS = 49152

WS_BORDER = 8388608
WS_CAPTION = 12582912
WS_CHILD = 1073741824
WS_CLIPCHILDREN = 33554432
WS_CLIPSIBLINGS = 67108864
WS_DISABLED = 134217728
WS_DLGFRAME = 4194304
WS_GROUP = 131072
WS_HSCROLL = 1048576
WS_MAXIMIZE = 16777216
WS_MAXIMIZEBOX = 65536
WS_MINIMIZE = 536870912
WS_MINIMIZEBOX = 131072
WS_OVERLAPPED = 0
WS_OVERLAPPEDWINDOW = 13565952
WS_POPUP = -2147483648
WS_POPUPWINDOW = -2138570752
WS_SYSMENU = 524288
WS_TABSTOP = 65536
WS_THICKFRAME = 262144
WS_VISIBLE = 268435456
WS_VSCROLL = 2097152

cRECT_3DDOWN = 2
cRECT_3DDOWN_FLAT = 6
cRECT_3DDOWN_SOLID = 3
cRECT_3DUP = 4
cRECT_3DUP_FLAT = 7
cRECT_3DUP_SOLID = 5
cRECT_BLACK_FRAME = 8
cRECT_BLACK_SOLID = 10
cRECT_BLACK_THICK_FRAME = 9
cRECT_ELLIPSE = 1
cRECT_FOCUS = 11
cRECT_GRAY_FRAME = 13
cRECT_GRAY_SOLID = 14
cRECT_INVERT = 19
cRECT_LTGRAY_SOLID = 15
cRECT_RECTANGLE = 0
cRECT_SIZE = 12
cRECT_WHITE_SOLID = 16
cRECT_YELLOW_FRAME = 17
cRECT_YELLOW_SOLID = 18
